"""Module for functional data manipulation.

Defines the abstract class that should be implemented by the funtional data
objects of the package and contains some commons methods.
"""

from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Iterator,
    NoReturn,
    Optional,
    Sequence,
    TypeVar,
    Union,
    cast,
    overload,
)

import numpy as np
import pandas.api.extensions
from typing_extensions import Literal

from .._utils import _evaluate_grid, _reshape_eval_points, _to_grid_points
from ._typing import (
    ArrayLike,
    DomainRange,
    GridPointsLike,
    LabelTuple,
    LabelTupleLike,
    NDArrayInt,
)
from .evaluator import Evaluator
from .extrapolation import ExtrapolationLike, _parse_extrapolation

if TYPE_CHECKING:
    from . import FDataGrid, FDataBasis
    from .basis import Basis

T = TypeVar('T', bound='FData')

EvalPointsType = Union[
    ArrayLike,
    Iterable[ArrayLike],
    GridPointsLike,
    Iterable[GridPointsLike],
]


class FData(  # noqa: WPS214
    ABC,
    pandas.api.extensions.ExtensionArray,  # type: ignore
):
    """Defines the structure of a functional data object.

    Attributes:
        n_samples (int): Number of samples.
        dim_domain (int): Dimension of the domain.
        dim_codomain (int): Dimension of the image.
        extrapolation (Extrapolation): Default extrapolation mode.
        dataset_name (str): name of the dataset.
        argument_names (tuple): tuple containing the names of the different
            arguments.
        coordinate_names (tuple): tuple containing the names of the different
            coordinate functions.

    """

    def __init__(
        self,
        *,
        extrapolation: Optional[ExtrapolationLike] = None,
        dataset_name: Optional[str] = None,
        dataset_label: Optional[str] = None,
        axes_labels: Optional[LabelTupleLike] = None,
        argument_names: Optional[LabelTupleLike] = None,
        coordinate_names: Optional[LabelTupleLike] = None,
        sample_names: Optional[LabelTupleLike] = None,
    ) -> None:

        self.extrapolation = extrapolation  # type: ignore
        self.dataset_name = dataset_name

        if dataset_label is not None:
            self.dataset_label = dataset_label

        self.argument_names = argument_names  # type: ignore
        self.coordinate_names = coordinate_names  # type: ignore
        if axes_labels is not None:
            self.axes_labels = axes_labels  # type: ignore
        self.sample_names = sample_names  # type: ignore

    @property
    def dataset_label(self) -> Optional[str]:
        warnings.warn(
            "Parameter dataset_label is deprecated. Use the "
            "parameter dataset_name instead.",
            DeprecationWarning,
        )
        return self.dataset_name

    @dataset_label.setter
    def dataset_label(self, name: Optional[str]) -> None:
        warnings.warn(
            "Parameter dataset_label is deprecated. Use the "
            "parameter dataset_name instead.",
            DeprecationWarning,
        )
        self.dataset_name = name

    @property
    def argument_names(self) -> LabelTuple:
        return self._argument_names

    @argument_names.setter
    def argument_names(
        self,
        names: Optional[LabelTupleLike],
    ) -> None:
        if names is None:
            names = (None,) * self.dim_domain
        else:
            names = tuple(names)
            if len(names) != self.dim_domain:
                raise ValueError(
                    "There must be a name for each of the "
                    "dimensions of the domain.",
                )

        self._argument_names = names

    @property
    def coordinate_names(self) -> LabelTuple:
        return self._coordinate_names

    @coordinate_names.setter
    def coordinate_names(
        self,
        names: Optional[LabelTupleLike],
    ) -> None:
        if names is None:
            names = (None,) * self.dim_codomain
        else:
            names = tuple(names)
            if len(names) != self.dim_codomain:
                raise ValueError(
                    "There must be a name for each of the "
                    "dimensions of the codomain.",
                )

        self._coordinate_names = names

    @property
    def axes_labels(self) -> LabelTuple:
        warnings.warn(
            "Parameter axes_labels is deprecated. Use the "
            "parameters argument_names and "
            "coordinate_names instead.",
            DeprecationWarning,
        )

        return self.argument_names + self.coordinate_names

    @axes_labels.setter
    def axes_labels(self, labels: LabelTupleLike) -> None:
        """Set the list of labels."""
        if labels is not None:

            warnings.warn(
                "Parameter axes_labels is deprecated. Use the "
                "parameters argument_names and "
                "coordinate_names instead.",
                DeprecationWarning,
            )

            labels_array = np.asarray(labels)
            if len(labels_array) > (self.dim_domain + self.dim_codomain):
                raise ValueError(
                    "There must be a label for each of the "
                    "dimensions of the domain and the image.",
                )
            if len(labels_array) < (self.dim_domain + self.dim_codomain):
                diff = (
                    (self.dim_domain + self.dim_codomain)
                    - len(labels_array)
                )
                labels_array = np.concatenate((labels_array, diff * [None]))

            self.argument_names = labels_array[:self.dim_domain]
            self.coordinate_names = labels_array[self.dim_domain:]

    @property
    def sample_names(self) -> LabelTuple:
        return self._sample_names

    @sample_names.setter
    def sample_names(self, names: Optional[LabelTupleLike]) -> None:
        if names is None:
            names = (None,) * self.n_samples
        else:
            names = tuple(names)
            if len(names) != self.n_samples:
                raise ValueError(
                    "There must be a name for each of the samples.",
                )

        self._sample_names = names

    @property
    @abstractmethod
    def n_samples(self) -> int:
        """Return the number of samples.

        Returns:
            Number of samples of the FData object.

        """
        pass

    @property
    @abstractmethod
    def dim_domain(self) -> int:
        """Return number of dimensions of the :term:`domain`.

        Returns:
            Number of dimensions of the domain.

        """
        pass

    @property
    @abstractmethod
    def dim_codomain(self) -> int:
        """Return number of dimensions of the :term:`codomain`.

        Returns:
            Number of dimensions of the codomain.

        """
        pass

    @property
    @abstractmethod
    def coordinates(self: T) -> Sequence[T]:
        r"""Return a component of the FDataGrid.

        If the functional object contains multivariate samples
        :math:`f: \mathbb{R}^n \rightarrow \mathbb{R}^d`, this method returns
        an iterator of the vector :math:`f = (f_1, ..., f_d)`.

        """
        pass

    @property
    def extrapolation(self) -> Optional[Evaluator]:
        """Return default type of extrapolation."""
        return self._extrapolation

    @extrapolation.setter
    def extrapolation(self, value: Optional[ExtrapolationLike]) -> None:
        """Set the type of extrapolation."""
        self._extrapolation = _parse_extrapolation(value)

    @property
    @abstractmethod
    def domain_range(self) -> DomainRange:
        """Return the :term:`domain` range of the object

        Returns:
            List of tuples with the ranges for each domain dimension.
        """
        pass

    def _extrapolation_index(self, eval_points: np.ndarray) -> np.ndarray:
        """Check the points that need to be extrapolated.

        Args:
            eval_points: Array with shape `n_eval_points` x
                `dim_domain` with the evaluation points, or shape ´n_samples´ x
                `n_eval_points` x `dim_domain` with different evaluation
                points for each sample.

        Returns:
            Array with boolean index. The positions with True
            in the index are outside the domain range and extrapolation
            should be applied.

        """
        index = np.zeros(eval_points.shape[:-1], dtype=np.bool_)

        # Checks bounds in each domain dimension
        for i, bounds in enumerate(self.domain_range):
            np.logical_or(index, eval_points[..., i] < bounds[0], out=index)
            np.logical_or(index, eval_points[..., i] > bounds[1], out=index)

        return index

    def _join_evaluation(
        self,
        index_matrix: np.ndarray,
        index_ext: np.ndarray,
        index_ev: np.ndarray,
        res_extrapolation: np.ndarray,
        res_evaluation: np.ndarray,
    ) -> np.ndarray:
        """Join the points evaluated.

        This method is used internally by :func:`evaluate` to join the result
        of the evaluation and the result of the extrapolation.

        Args:
            index_matrix: Boolean index with the points extrapolated.
            index_ext: Boolean index with the columns that contains
                points extrapolated.
            index_ev: Boolean index with the columns that contains
                points evaluated.
            res_extrapolation: Result of the extrapolation.
            res_evaluation: Result of the evaluation.

        Returns:
            Matrix with the points evaluated with shape
            `n_samples` x `number of points evaluated` x `dim_codomain`.

        """
        res = np.empty((
            self.n_samples,
            index_matrix.shape[-1],
            self.dim_codomain,
        ))

        # Case aligned evaluation
        if index_matrix.ndim == 1:
            res[:, index_ev, :] = res_evaluation
            res[:, index_ext, :] = res_extrapolation

        else:

            res[:, index_ev] = res_evaluation
            res[index_matrix] = res_extrapolation[index_matrix[:, index_ext]]

        return res

    @abstractmethod
    def _evaluate(
        self,
        eval_points: Union[ArrayLike, Iterable[ArrayLike]],
        *,
        aligned: bool = True,
    ) -> np.ndarray:
        """Define the evaluation of the FData.

        Evaluates the samples of an FData object at several points.

        Subclasses must override this method to implement evaluation.

        Args:
            eval_points: List of points where the functions are
                evaluated. If `aligned` is `True`, then a list of
                lists of points must be passed, with one list per sample.
            aligned: Whether the input points are
                the same for each sample, or an array of points per sample is
                passed.

        Returns:
            Numpy 3d array with shape `(n_samples,
            len(eval_points), dim_codomain)` with the result of the
            evaluation. The entry (i,j,k) will contain the value k-th image
            dimension of the i-th sample, at the j-th evaluation point.

        """
        pass

    @overload
    def evaluate(
        self,
        eval_points: ArrayLike,
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[False] = False,
        aligned: Literal[True] = True,
    ) -> np.ndarray:
        pass

    @overload
    def evaluate(
        self,
        eval_points: Iterable[ArrayLike],
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[False] = False,
        aligned: Literal[False],
    ) -> np.ndarray:
        pass

    @overload
    def evaluate(
        self,
        eval_points: GridPointsLike,
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[True],
        aligned: Literal[True] = True,
    ) -> np.ndarray:
        pass

    @overload
    def evaluate(
        self,
        eval_points: Iterable[GridPointsLike],
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[True],
        aligned: Literal[False],
    ) -> np.ndarray:
        pass

    def evaluate(
        self,
        eval_points: EvalPointsType,
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: bool = False,
        aligned: bool = True,
    ) -> np.ndarray:
        """Evaluate the object at a list of values or a grid.

        Args:
            eval_points: List of points where the functions are
                evaluated. If ``grid`` is ``True``, a list of axes, one per
                :term:`domain` dimension, must be passed instead. If
                ``aligned`` is ``True``, then a list of lists (of points or
                axes, as explained) must be passed, with one list per sample.
            derivative: Deprecated. Order of the derivative to evaluate.
            extrapolation: Controls the
                extrapolation mode for elements outside the domain range. By
                default it is used the mode defined during the instance of the
                object.
            grid: Whether to evaluate the results on a grid
                spanned by the input arrays, or at points specified by the
                input arrays. If true the eval_points should be a list of size
                dim_domain with the corresponding times for each axis. The
                return matrix has shape n_samples x len(t1) x len(t2) x ... x
                len(t_dim_domain) x dim_codomain. If the domain dimension is 1
                the parameter has no efect. Defaults to False.
            aligned: Whether the input points are the same for each sample,
                or an array of points per sample is passed.

        Returns:
            Matrix whose rows are the values of the each
            function at the values specified in eval_points.

        """
        if derivative != 0:
            warnings.warn(
                "Parameter derivative is deprecated. Use the "
                "derivative function instead.",
                DeprecationWarning,
            )
            return self.derivative(order=derivative)(  # type: ignore
                eval_points,
                extrapolation=extrapolation,
                grid=grid,
                aligned=aligned,
            )

        if grid:  # Evaluation of a grid performed in auxiliar function

            return _evaluate_grid(  # type: ignore
                eval_points,
                evaluate_method=self.evaluate,
                n_samples=self.n_samples,
                dim_domain=self.dim_domain,
                dim_codomain=self.dim_codomain,
                extrapolation=extrapolation,
                aligned=aligned,
            )

        eval_points = cast(Union[ArrayLike, Iterable[ArrayLike]], eval_points)

        if extrapolation is None:
            extrapolation = self.extrapolation
        else:
            # Gets the function to perform extrapolation or None
            extrapolation = _parse_extrapolation(extrapolation)

        eval_points = cast(
            Union[ArrayLike, Sequence[ArrayLike]],
            eval_points,
        )

        # Convert to array and check dimensions of eval points
        eval_points = _reshape_eval_points(
            eval_points,
            aligned=aligned,
            n_samples=self.n_samples,
            dim_domain=self.dim_domain,
        )

        if extrapolation is not None:

            index_matrix = self._extrapolation_index(eval_points)

            if index_matrix.any():

                # Partition of eval points
                if aligned:

                    index_ext = index_matrix
                    index_ev = ~index_matrix

                    eval_points_extrapolation = eval_points[index_ext]
                    eval_points_evaluation = eval_points[index_ev]

                else:
                    index_ext = np.logical_or.reduce(index_matrix, axis=0)
                    eval_points_extrapolation = eval_points[:, index_ext]

                    index_ev = np.logical_or.reduce(~index_matrix, axis=0)
                    eval_points_evaluation = eval_points[:, index_ev]

                # Direct evaluation
                res_evaluation = self._evaluate(
                    eval_points_evaluation,
                    aligned=aligned,
                )

                res_extrapolation = extrapolation(  # type: ignore
                    self,
                    eval_points_extrapolation,
                    aligned=aligned,
                )

                return self._join_evaluation(
                    index_matrix,
                    index_ext,
                    index_ev,
                    res_extrapolation,
                    res_evaluation,
                )

        return self._evaluate(
            eval_points,
            aligned=aligned,
        )

    @overload
    def __call__(
        self,
        eval_points: ArrayLike,
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[False] = False,
        aligned: Literal[True] = True,
    ) -> np.ndarray:
        pass

    @overload
    def __call__(
        self,
        eval_points: Iterable[ArrayLike],
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[False] = False,
        aligned: Literal[False],
    ) -> np.ndarray:
        pass

    @overload
    def __call__(
        self,
        eval_points: GridPointsLike,
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[True],
        aligned: Literal[True] = True,
    ) -> np.ndarray:
        pass

    @overload
    def __call__(
        self,
        eval_points: Iterable[GridPointsLike],
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[True],
        aligned: Literal[False],
    ) -> np.ndarray:
        pass

    def __call__(
        self,
        eval_points: EvalPointsType,
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: bool = False,
        aligned: bool = True,
    ) -> np.ndarray:
        """Evaluate the :term:`functional object`.

        Evaluate the object or its derivatives at a list of values or a
        grid. This method is a wrapper of :meth:`evaluate`.

        Args:
            eval_points: List of points where the functions are
                evaluated. If a matrix of shape nsample x eval_points is given
                each sample is evaluated at the values in the corresponding row
                in eval_points.
            derivative: Order of the derivative. Defaults to 0.
            extrapolation: Controls the
                extrapolation mode for elements outside the domain range. By
                default it is used the mode defined during the instance of the
                object.
            grid: Whether to evaluate the results on a grid
                spanned by the input arrays, or at points specified by the
                input arrays. If true the eval_points should be a list of size
                dim_domain with the corresponding times for each axis. The
                return matrix has shape n_samples x len(t1) x len(t2) x ... x
                len(t_dim_domain) x dim_codomain. If the domain dimension is 1
                the parameter has no efect. Defaults to False.
            aligned: Whether the input points are the same for each sample,
                or an array of points per sample is passed.

        Returns:
            Matrix whose rows are the values of the each
            function at the values specified in eval_points.

        """
        return self.evaluate(  # type: ignore
            eval_points,
            derivative=derivative,
            extrapolation=extrapolation,
            grid=grid,
            aligned=aligned,
        )

    @abstractmethod
    def derivative(self: T, *, order: int = 1) -> T:
        """Differentiate a FData object.

        Args:
            order: Order of the derivative. Defaults to one.

        Returns:
            Functional object containg the derivative.

        """
        pass

    @abstractmethod
    def shift(
        self,
        shifts: Union[ArrayLike, float],
        *,
        restrict_domain: bool = False,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid_points: Optional[GridPointsLike] = None,
    ) -> FDataGrid:
        r"""
        Perform a shift of the curves.

        The i-th shifted function :math:`y_i` has the form

        .. math::
            y_i(t) = x_i(t + \delta_i)

        where :math:`x_i` is the i-th original function and :math:`delta_i` is
        the shift performed for that function, that must be a vector in the
        domain space.

        Note that a positive shift moves the graph of the function in the
        negative direction and vice versa.

        Args:
            shifts: List with the shifts
                corresponding for each sample or numeric with the shift to
                apply to all samples.
            restrict_domain: If True restricts the domain to avoid the
                evaluation of points outside the domain using extrapolation.
                Defaults uses extrapolation.
            extrapolation: Controls the
                extrapolation mode for elements outside the domain range.
                By default uses the method defined in fd. See extrapolation to
                more information.
            grid_points: Grid of points where
                the functions are evaluated to obtain the discrete
                representation of the object to operate. If ``None`` the
                current grid_points are used to unificate the domain of the
                shifted data.

        Returns:
            Shifted functions.

        """
        assert grid_points is not None
        grid_points = _to_grid_points(grid_points)

        arr_shifts = np.array([shifts] if np.isscalar(shifts) else shifts)

        # Accept unidimensional array when the domain dimension is one or when
        # the shift is the same for each sample
        if arr_shifts.ndim == 1:
            arr_shifts = (
                arr_shifts[np.newaxis, :]  # Same shift for each sample
                if len(arr_shifts) == self.dim_domain
                else arr_shifts[:, np.newaxis]
            )

        if len(arr_shifts) not in {1, self.n_samples}:
            raise ValueError(
                f"The length of the shift vector ({len(arr_shifts)}) must "
                f"have length equal to 1 or to the number of samples "
                f"({self.n_samples})",
            )

        if restrict_domain:
            domain = np.asarray(self.domain_range)

            a = domain[:, 0] - np.min(np.min(arr_shifts, axis=0), 0)
            b = domain[:, 1] - np.max(np.max(arr_shifts, axis=1), 0)

            domain = np.hstack((a, b))
            domain_range = tuple(domain)

        if len(arr_shifts) == 1:
            shifted_grid_points = tuple(
                g + s for g, s in zip(grid_points, arr_shifts[0])
            )
            data_matrix = self(
                shifted_grid_points,
                extrapolation=extrapolation,
                aligned=True,
                grid=True,
            )
        else:
            shifted_grid_points_per_sample = (
                tuple(
                    g + s for g, s in zip(grid_points, shift)
                ) for shift in arr_shifts
            )
            data_matrix = self(
                shifted_grid_points_per_sample,
                extrapolation=extrapolation,
                aligned=False,
                grid=True,
            )

        shifted = self.to_grid().copy(
            data_matrix=data_matrix,
            grid_points=grid_points,
        )

        if restrict_domain:
            shifted = shifted.restrict(domain_range)

        return shifted

    def plot(self, *args: Any, **kwargs: Any) -> Any:
        """Plot the FDatGrid object.

        Args:
            args: positional arguments for :func:`plot_graph`.
            kwargs: keyword arguments for :func:`plot_graph`.

        Returns:
            fig (figure object): figure object in which the graphs are plotted.

        """
        from ..exploratory.visualization.representation import GraphPlot

        return GraphPlot(self, *args, **kwargs).plot()

    @abstractmethod
    def copy(
        self: T,
        *,
        deep: bool = False,  # For Pandas compatibility
        dataset_name: Optional[str] = None,
        argument_names: Optional[LabelTupleLike] = None,
        coordinate_names: Optional[LabelTupleLike] = None,
        sample_names: Optional[LabelTupleLike] = None,
        extrapolation: Optional[ExtrapolationLike] = None,
    ) -> T:
        """Make a copy of the object."""
        pass

    @abstractmethod  # noqa: WPS125
    def sum(  # noqa: WPS125
        self: T,
        *,
        axis: Optional[int] = None,
        out: None = None,
        keepdims: bool = False,
        skipna: bool = False,
        min_count: int = 0,
    ) -> T:
        """Compute the sum of all the samples.

        Args:
            axis: Used for compatibility with numpy. Must be None or 0.
            out: Used for compatibility with numpy. Must be None.
            keepdims: Used for compatibility with numpy. Must be False.
            skipna: Wether the NaNs are ignored or not.
            min_count: Number of valid (non NaN) data to have in order
                for the a variable to not be NaN when `skipna` is
                `True`.

        Returns:
            A FData object with just one sample representing
            the sum of all the samples in the original object.

        """
        if (
            (axis is not None and axis != 0)
            or out is not None
            or keepdims is not False
        ):
            raise NotImplementedError(
                "Not implemented for that parameter combination",
            )

        return self

    def mean(
        self: T,
        *,
        axis: None = None,
        dtype: None = None,
        out: None = None,
        keepdims: bool = False,
        skipna: bool = False,
    ) -> T:
        """Compute the mean of all the samples.

        Args:
            axis: Used for compatibility with numpy. Must be None or 0.
            dtype: Used for compatibility with numpy. Must be None.
            out: Used for compatibility with numpy. Must be None.
            keepdims: Used for compatibility with numpy. Must be False.
            skipna: Wether the NaNs are ignored or not.

        Returns:
            A FData object with just one sample representing
            the mean of all the samples in the original object.

        """
        if dtype is not None:
            raise NotImplementedError(
                "Not implemented for that parameter combination",
            )

        return (
            self.sum(axis=axis, out=out, keepdims=keepdims, skipna=skipna)
            / self.n_samples
        )

    @abstractmethod
    def to_grid(
        self,
        grid_points: Optional[GridPointsLike] = None,
    ) -> FDataGrid:
        """Return the discrete representation of the object.

        Args:
            grid_points: Points per axis
                where the function is going to be evaluated.

        Returns:
            Discrete representation of the functional data
            object.

        """
        pass

    @abstractmethod
    def to_basis(
        self,
        basis: Basis,
        **kwargs: Any,
    ) -> FDataBasis:
        """Return the basis representation of the object.

        Args:
            basis: basis object in which the functional data are
                going to be represented.
            kwargs: keyword arguments to be passed to
                FDataBasis.from_data().

        Returns:
            Basis representation of the funtional data
            object.

        """
        pass

    @abstractmethod
    def concatenate(self: T, *others: T, as_coordinates: bool = False) -> T:
        """Join samples from a similar FData object.

        Joins samples from another FData object if it has the same
        dimensions and has compatible representations.

        Args:
            others: other FData objects.
            as_coordinates:  If False concatenates as
                new samples, else, concatenates the other functions as
                new components of the image. Defaults to False.

        Returns:
            :class:`FData`: FData object with the samples from the two
            original objects.

        """
        pass

    @abstractmethod
    def compose(
        self: T,
        fd: T,
        *,
        eval_points: Optional[np.ndarray] = None,
    ) -> FData:
        """Composition of functions.

        Performs the composition of functions.

        Args:
            fd: FData object to make the composition. Should
                have the same number of samples and image dimension equal to
                the domain dimension of the object composed.
            eval_points: Points to perform the evaluation.

        """
        pass

    @abstractmethod
    def __getitem__(self: T, key: Union[int, slice, np.ndarray]) -> T:
        """Return self[key]."""
        pass

    def equals(self, other: object) -> bool:
        """Whole object equality."""
        return (
            isinstance(other, type(self))  # noqa: WPS222
            and self.extrapolation == other.extrapolation
            and self.dataset_name == other.dataset_name
            and self.argument_names == other.argument_names
            and self.coordinate_names == other.coordinate_names
        )

    @abstractmethod
    def __eq__(self, other: object) -> np.ndarray:  # type: ignore[override]
        pass

    def __ne__(self, other: object) -> np.ndarray:  # type: ignore[override]
        """Return for `self != other` (element-wise in-equality)."""
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented

        return ~result

    def _copy_op(
        self: T,
        other: Union[T, np.ndarray, float],
        **kwargs: Any,
    ) -> T:

        base_copy = (
            other if isinstance(other, type(self))
            and self.n_samples == 1 and other.n_samples != 1
            else self
        )

        return base_copy.copy(**kwargs)

    @abstractmethod
    def __add__(self: T, other: Union[T, np.ndarray, float]) -> T:
        """Addition for FData object."""
        pass

    @abstractmethod
    def __radd__(self: T, other: Union[T, np.ndarray, float]) -> T:
        """Addition for FData object."""
        pass

    @abstractmethod
    def __sub__(self: T, other: Union[T, np.ndarray, float]) -> T:
        """Subtraction for FData object."""
        pass

    @abstractmethod
    def __rsub__(self: T, other: Union[T, np.ndarray, float]) -> T:
        """Right subtraction for FData object."""
        pass

    @abstractmethod
    def __mul__(self: T, other: Union[np.ndarray, float]) -> T:
        """Multiplication for FData object."""
        pass

    @abstractmethod
    def __rmul__(self: T, other: Union[np.ndarray, float]) -> T:
        """Multiplication for FData object."""
        pass

    @abstractmethod
    def __truediv__(self: T, other: Union[np.ndarray, float]) -> T:
        """Division for FData object."""
        pass

    @abstractmethod
    def __rtruediv__(self: T, other: Union[np.ndarray, float]) -> T:
        """Right division for FData object."""
        pass

    def __iter__(self: T) -> Iterator[T]:
        """Iterate over the samples."""
        yield from (self[i] for i in range(self.n_samples))

    def __len__(self) -> int:
        """Return the number of samples of the FData object."""
        return self.n_samples

    #####################################################################
    # Numpy methods
    #####################################################################

    def __array__(self, *args: Any, **kwargs: Any) -> np.ndarray:
        """Return a numpy array with the objects."""
        # This is to prevent numpy to access inner dimensions
        array = np.empty(shape=len(self), dtype=np.object_)

        for i, f in enumerate(self):
            array[i] = f

        return array

    def __array_ufunc__(
        self,
        ufunc: Any,
        method: str,
        *inputs: Any,
        **kwargs: Any,
    ) -> Any:
        """Prevent NumPy from converting to array just to do operations."""

        # Make normal multiplication by scalar use the __mul__ method
        if ufunc == np.multiply and method == "__call__" and len(inputs) == 2:
            if isinstance(inputs[0], np.ndarray):
                inputs = inputs[::-1]

            return inputs[0] * inputs[1]

        return NotImplemented

    #####################################################################
    # Pandas ExtensionArray methods
    #####################################################################
    @property
    def ndim(self) -> int:
        """
        Return number of dimensions of the functional data.

        It is always 1, as each observation is considered a "scalar" object.

        Returns:
            Number of dimensions of the functional data.

        """
        return 1

    @classmethod
    def _from_sequence(
        cls,
        scalars: Union['FData', Sequence['FData']],
        dtype: Any = None,
        copy: bool = False,
    ) -> 'FData':

        scalars_seq: Sequence['FData'] = (
            [scalars] if isinstance(scalars, cls) else scalars
        )

        if copy:
            scalars_seq = [f.copy() for f in scalars_seq]

        if dtype is None:
            first_element = next(s for s in scalars_seq if s is not pandas.NA)
            dtype = first_element.dtype

        scalars_seq = [
            s if s is not pandas.NA else dtype._na_repr()  # noqa: WPS437
            for s in scalars_seq
        ]

        if len(scalars_seq) == 0:
            scalars_seq = [dtype._na_repr()[:0]]  # noqa: WPS437

        return cls._concat_same_type(scalars_seq)

    @classmethod
    def _from_factorized(cls, values: Any, original: Any) -> NoReturn:
        raise NotImplementedError(
            "Factorization does not make sense for functional data",
        )

    @abstractmethod
    def _take_allow_fill(
        self: T,
        indices: NDArrayInt,
        fill_value: T,
    ) -> T:
        pass

    def take(
        self: T,
        indices: Union[int, Sequence[int], NDArrayInt],
        allow_fill: bool = False,
        fill_value: Optional[T] = None,
        axis: int = 0,
    ) -> T:
        """Take elements from an array.

        Parameters:
            indices:
                Indices to be taken.
            allow_fill: How to handle negative values in `indices`.
                * False: negative values in `indices` indicate positional
                  indices from the right (the default). This is similar to
                  :func:`numpy.take`.
                * True: negative values in `indices` indicate
                  missing values. These values are set to `fill_value`. Any
                  other negative values raise a ``ValueError``.
            fill_value: Fill value to use for NA-indices
                when `allow_fill` is True.
                This may be ``None``, in which case the default NA value for
                the type, ``self.dtype.na_value``, is used.
                For many ExtensionArrays, there will be two representations of
                `fill_value`: a user-facing "boxed" scalar, and a low-level
                physical NA value. `fill_value` should be the user-facing
                version, and the implementation should handle translating that
                to the physical version for processing the take if necessary.
            axis: Parameter for compatibility with numpy. Must be always 0.

        Returns:
            FData

        Raises:
            IndexError: When the indices are out of bounds for the array.
            ValueError: When `indices` contains negative values other than
                ``-1`` and `allow_fill` is True.

        Notes:
            ExtensionArray.take is called by ``Series.__getitem__``, ``.loc``,
            ``iloc``, when `indices` is a sequence of values. Additionally,
            it's called by :meth:`Series.reindex`, or any other method
            that causes realignment, with a `fill_value`.

        See Also:
            numpy.take
            pandas.api.extensions.take
        """
        # The axis parameter must exist, because sklearn tries to use take
        # instead of __getitem__
        if axis != 0:
            raise ValueError(f"Axis must be 0, not {axis}")

        arr_indices = np.atleast_1d(indices)

        if fill_value is None:
            fill_value = self.dtype.na_value

        non_empty_take_msg = "cannot do a non-empty take from an empty axes"

        if allow_fill:
            if (arr_indices < -1).any():
                raise ValueError("Invalid indexes")

            positive_mask = arr_indices >= 0
            if len(self) == 0 and positive_mask.any():
                raise IndexError(non_empty_take_msg)

            sample_names = np.zeros(len(arr_indices), dtype=object)
            result = self._take_allow_fill(arr_indices, fill_value)

            sample_names[positive_mask] = np.array(self.sample_names)[
                arr_indices[positive_mask]
            ]

            if fill_value is not self.dtype.na_value:
                sample_names[~positive_mask] = fill_value.sample_names[0]

            result.sample_names = tuple(sample_names)
        else:
            if len(self) == 0 and len(arr_indices) != 0:
                raise IndexError(non_empty_take_msg)

            result = self[arr_indices]

        return result

    @classmethod
    def _concat_same_type(
        cls,
        to_concat: Sequence[T],
    ) -> T:
        """
        Concatenate multiple array.

        Parameters:
            to_concat: Sequence of FData objects to concat.

        Returns:
            Concatenation of the objects.

        """
        if isinstance(to_concat, cls):
            return to_concat

        return concatenate(to_concat)

    def astype(self, dtype: Any, copy: bool = True) -> Any:
        """Cast to a new dtype."""
        if isinstance(dtype, type(self.dtype)):
            new_obj = self
            if copy:
                new_obj = self.copy()
            return new_obj

        return super().astype(dtype)

    def _reduce(self, name: str, skipna: bool = True, **kwargs: Any) -> Any:
        meth = getattr(self, name, None)
        if meth:
            return meth(skipna=skipna, **kwargs)

        raise TypeError(
            f"'{type(self).__name__}' does not implement "
            f"reduction '{name}'",
        )


def concatenate(functions: Iterable[T], as_coordinates: bool = False) -> T:
    """
    Join samples from an iterable of similar FData objects.

    Joins samples of FData objects if they have the same
    dimensions and sampling points.
    Args:
        objects: Objects to be concatenated.
        as_coordinates:  If False concatenates as
                new samples, else, concatenates the other functions as
                new components of the image. Defaults to False.
    Returns:
        FData object with the samples from the
        original objects.
    Raises:
        ValueError: In case the provided list of FData objects is
        empty.
    Todo:
        By the moment, only unidimensional objects are supported in basis
        representation.
    """
    functions = iter(functions)
    first = next(functions, None)

    if first is None:
        raise ValueError(
            "At least one FData object must be provided to concatenate.",
        )

    return first.concatenate(*functions, as_coordinates=as_coordinates)



import abc
import copy
import numbers
import random

import numpy as np
import numpy.linalg as linalg
import numpy.ma as ma
import scipy.stats
import sklearn.base
import sklearn.utils

import dcor

from ...._utils import _compute_dependence
from ....representation import FDataGrid


def _transform_to_2d(t):
    t = np.asarray(t)

    dim = len(t.shape)
    assert dim <= 2

    if dim < 2:
        t = np.atleast_2d(t).T

    return t


def _execute_kernel(kernel, t_0, t_1):
    from ....misc.covariances import _execute_covariance

    return _execute_covariance(kernel, t_0, t_1)


class _PicklableKernel():
    """ Class used to pickle GPy kernels."""

    def __init__(self, kernel):
        super().__setattr__('_PicklableKernel__kernel', kernel)

    def __getattr__(self, name):
        if name != '__deepcopy__':
            return getattr(self.__kernel, name)

    def __setattr__(self, name, value):
        setattr(self.__kernel, name, value)

    def __getstate__(self):
        return {'class': self.__kernel.__class__,
                'input_dim': self.__kernel.input_dim,
                'values': self.__kernel.param_array}

    def __setstate__(self, state):
        super().__setattr__('_PicklableKernel__kernel', state['class'](
            input_dim=state['input_dim']))
        self.__kernel.param_array[...] = state['values']

    def __call__(self, *args, **kwargs):
        return self.__kernel.K(*args, **kwargs)


def make_kernel(k):
    try:
        import GPy
    except ImportError:
        return k

    if isinstance(k, GPy.kern.Kern):
        return _PicklableKernel(k)
    else:
        return k


def _absolute_argmax(function, *, mask):
    """
    Computes the absolute maximum of a discretized function.

    Some values of the function may be masked in order not to consider them
    as maximum.

    Parameters:
        function (numpy array): Discretized function.
        mask (numpy boolean array): Masked values.

    Returns:
        int: Index of the absolute maximum.

    """
    masked_function = ma.array(function, mask=mask)

    t_max = ma.argmax(masked_function)

    t_max = np.unravel_index(t_max, function.shape)

    return t_max


class Correction(abc.ABC, sklearn.base.BaseEstimator):
    """
    Base class for applying a correction after a point is taken, eliminating
    its influence over the rest.

    """

    def begin(self, X: FDataGrid, Y):
        """
        Initialization for a particular application of Recursive Maxima
        Hunting.

        The initial parameters of Recursive Maxima Hunting can be used there.

        """
        pass

    def conditioned(self, **kwargs):
        """
        Returns a correction object conditioned to the value of a point.

        This method is necessary because after the RMH correction step, the
        functions follow a different model.

        """
        return self

    @abc.abstractmethod
    def correct(self, X, selected_index):
        """
        Correct the trajectories.

        This method subtracts the influence of the selected point from the
        other points in the function.

        Parameters:

            X (FDataGrid): Functions in the current iteration of the algorithm.
            selected_index (int or tuple of int): Index of the selected point
                in the ``data_matrix``.

        """
        pass

    def __call__(self, *args, **kwargs):
        self.correct(*args, **kwargs)


class ConditionalMeanCorrection(Correction):
    """
    Base class for applying a correction based on the conditional expectation.

    The functions are assumed to be realizations of a particular stochastic
    process. The information subtracted in each iteration would be the
    mean of the process conditioned to the value observed at the
    selected point.

    """

    @abc.abstractmethod
    def conditional_mean(self, X, selected_index):
        """
        Mean of the process conditioned to the value observed at the
        selected point.

        Parameters:

            X (FDataGrid): Functions in the current iteration of the algorithm.
            selected_index (int or tuple of int): Index of the selected point
                in the ``data_matrix``.

        """
        pass

    def correct(self, X, selected_index):

        X.data_matrix[...] -= self.conditional_mean(
            X, selected_index).T

        X.data_matrix[:, selected_index] = 0


class GaussianCorrection(ConditionalMeanCorrection):
    r"""
    Correction assuming that the underlying process is Gaussian.

    The conditional mean of a Gaussian process :math:`X(t)` is

    .. math::

        \mathbb{E}[X(t) \mid X(t_0) = x_0] = \mathbb{E}[X(t)]
        + \frac{\mathrm{Cov}[X(t), X(t_0)]}{\mathrm{Cov}[X(t_0), X(t_0)]}
        (X(t_0) - \mathbb{E}[X(t_0)])

    The corrections after this is applied are of type
    :class:`GaussianConditionedCorrection`.

    Parameters:

        mean (number or function): Mean function of the Gaussian process.
        cov (number or function): Covariance function of the Gaussian process.
        fit_hyperparameters (boolean): If ``True`` the hyperparameters of the
            covariance function are optimized for the data.

    """

    def __init__(self, *, mean=0, cov=1, fit_hyperparameters=False):
        super(GaussianCorrection, self).__init__()

        self.mean = mean
        self.cov = make_kernel(cov)
        self.fit_hyperparameters = fit_hyperparameters

    def begin(self, X, y):
        if self.fit_hyperparameters:
            import GPy

            T = X.grid_points[0]
            X_copy = np.copy(X.data_matrix[..., 0])

            y = np.ravel(y)
            for class_label in np.unique(y):
                trajectories = X_copy[y == class_label, :]

                mean = np.mean(trajectories, axis=0)
                X_copy[y == class_label, :] -= mean

            m = GPy.models.GPRegression(
                T[:, None], X_copy.T,
                kernel=self.cov._PicklableKernel__kernel)
            m.constrain_positive('')
            m.optimize()

            self.cov_ = copy.deepcopy(make_kernel(m.kern))

    def _evaluate_mean(self, t):

        mean = self.mean

        if isinstance(mean, numbers.Number):
            expectation = np.ones_like(t, dtype=float) * mean
        else:
            expectation = mean(t)

        return expectation

    def _evaluate_cov(self, t_0, t_1):
        cov = getattr(self, "cov_", self.cov)

        return _execute_kernel(cov, t_0, t_1)

    def conditioned(self, X, t_0, **kwargs):
        # If the point makes the matrix singular, don't change the correction

        cov = getattr(self, "cov_", self.cov)

        try:

            correction = GaussianConditionedCorrection(
                mean=self.mean,
                cov=cov,
                conditioning_points=t_0)

            correction._covariance_matrix_inv()

            return correction

        except linalg.LinAlgError:

            return self

    def conditional_mean(self, X, selected_index):

        T = X.grid_points[0]

        t_0 = T[selected_index]

        x_index = (slice(None),) + tuple(selected_index) + (np.newaxis,)
        x_0 = X.data_matrix[x_index]

        T = _transform_to_2d(T)

        var = self._evaluate_cov(t_0, t_0)

        expectation = self._evaluate_mean(T)
        assert expectation.shape == T.shape

        t_0_expectation = expectation[selected_index]

        b_T = self._evaluate_cov(T, t_0)
        assert b_T.shape == T.shape

        cond_expectation = (expectation +
                            b_T / var *
                            (x_0.T - t_0_expectation)
                            ) if var else expectation + np.zeros_like(x_0.T)

        return cond_expectation


class GaussianConditionedCorrection(GaussianCorrection):
    r"""
    Correction assuming that the underlying process is Gaussian, with several
    values conditioned to 0.

   The conditional mean is inherited from :class:`GaussianCorrection`, with
   the conditioned mean and covariance.

    The corrections after this is applied are of type
    :class:`GaussianConditionedCorrection`, adding additional points.

    Parameters:

        conditioning_points (iterable of ints or tuples of ints): Points where
            the process is conditioned to have the value 0.
        mean (number or function): Mean function of the (unconditioned)
            Gaussian process.
        cov (number or function): Covariance function of the (unconditioned)
            Gaussian process.

    """

    def __init__(self, conditioning_points, *, mean=0, cov=1):

        super(GaussianConditionedCorrection, self).__init__(
            mean=mean, cov=cov)

        self.conditioning_points = conditioning_points

    def _covariance_matrix_inv(self):

        cond_points = self._conditioning_points()

        cov_matrix_inv = getattr(self, "_cov_matrix_inv", None)
        if cov_matrix_inv is None:

            cov_matrix = super()._evaluate_cov(
                cond_points, cond_points
            )

            self._cov_matrix_inv = np.linalg.inv(cov_matrix)
            cov_matrix_inv = self._cov_matrix_inv

        return cov_matrix_inv

    def _conditioning_points(self):
        return _transform_to_2d(self.conditioning_points)

    def conditioned(self, X, t_0, **kwargs):

        # If the point makes the matrix singular, don't change the correction
        try:

            correction = GaussianConditionedCorrection(
                mean=self.mean,
                cov=self.cov,
                conditioning_points=np.concatenate(
                    (self._conditioning_points(), [[t_0]]))
            )

            correction._covariance_matrix_inv()

            return correction

        except linalg.LinAlgError:

            return self

    def _evaluate_mean(self, t):

        cond_points = self._conditioning_points()

        A_inv = self._covariance_matrix_inv()

        b_T = super()._evaluate_cov(t, cond_points)

        c = -super()._evaluate_mean(cond_points)
        assert c.shape == np.shape(cond_points)

        original_expect = super()._evaluate_mean(t)
        assert original_expect.shape == t.shape

        modified_expect = b_T.dot(A_inv).dot(c)
        assert modified_expect.shape == t.shape

        expectation = original_expect + modified_expect
        assert expectation.shape == t.shape

        return expectation

    def _evaluate_cov(self, t_0, t_1):

        cond_points = self._conditioning_points()

        A_inv = self._covariance_matrix_inv()

        b_t_0_T = super()._evaluate_cov(t_0, cond_points)

        b_t_1 = super()._evaluate_cov(cond_points, t_1)

        return (super()._evaluate_cov(t_0, t_1) -
                b_t_0_T @ A_inv @ b_t_1)


class GaussianSampleCorrection(ConditionalMeanCorrection):
    """
    Correction assuming that the process is Gaussian and using as the kernel
    the sample covariance.

    """

    def begin(self, X: FDataGrid, Y):

        X_copy = np.copy(X.data_matrix[..., 0])

        Y = np.ravel(Y)
        for class_label in np.unique(Y):
            trajectories = X_copy[Y == class_label, :]

            mean = np.mean(trajectories, axis=0)
            X_copy[Y == class_label, :] -= mean

        self.cov_matrix_ = np.cov(X_copy, rowvar=False)
        self.t_ = np.ravel(X.grid_points)
        self.gaussian_correction_ = GaussianCorrection(
            cov=self.cov)

    def cov(self, t_0, t_1):
        i = np.searchsorted(self.t_, t_0)
        j = np.searchsorted(self.t_, t_1)

        i = np.ravel(i)
        j = np.ravel(j)

        return self.cov_matrix_[np.ix_(i, j)]

    def conditioned(self, t_0, **kwargs):
        self.gaussian_correction_ = self.gaussian_correction_.conditioned(
            t_0=t_0, **kwargs)
        return self

    def conditional_mean(self, X, selected_index):

        return self.gaussian_correction_.conditional_mean(
            X, selected_index)


class UniformCorrection(Correction):
    """
    Correction assuming that the underlying process is an Ornstein-Uhlenbeck
    process with infinite lengthscale.

    The initial conditional mean subtracts the observed value from every
    point, and the following correction is a :class:`GaussianCorrection`
    with a :class:`~skfda.misc.covariances.Brownian` covariance function with
    the selected point as its origin.

    """

    def conditioned(self, X, t_0, **kwargs):
        from ....misc.covariances import Brownian

        return GaussianCorrection(cov=Brownian(origin=t_0))

    def correct(self, X, selected_index):
        x_index = (slice(None),) + tuple(selected_index) + (np.newaxis,)

        # Have to copy it because otherwise is a view and shouldn't be
        # subtracted from the original matrix
        x_0 = np.copy(X.data_matrix[x_index])

        X.data_matrix[...] -= x_0


class StoppingCondition(abc.ABC, sklearn.base.BaseEstimator):
    """
    Stopping condition for RMH.

    This is a callable that should return ``True`` if the algorithm must stop
    and the current point should not be selected.

    """

    @abc.abstractmethod
    def __call__(self, **kwargs):
        pass


class ScoreThresholdStop(StoppingCondition):
    """
    Stop when the score is under a threshold.

    This stopping condition requires that the score has a known bound, for
    example that it takes values in the interval :math:`[0, 1]`.

    This is one of the simplest stopping criterions, but it requires that
    the user chose a threshold parameter, which controls the number of
    points chosen and can vary per problem.

    Parameters:

        threshold (float): Value compared with the score. If the score
                           of the selected point is not higher than that,
                           the point will not be selected (unless it is
                           the first iteration) and RMH will end.

    """

    def __init__(self, threshold=0.2):

        super().__init__()
        self.threshold = threshold

    def __call__(self, *, selected_index, dependences, **kwargs):

        score = dependences[selected_index]

        return score < self.threshold


class AsymptoticIndependenceTestStop(StoppingCondition):
    r"""
    Stop when the selected point is independent from the target.

    It uses an asymptotic test based on the chi-squared distribution described
    in :footcite:`szekely+rizzo_2010_brownian`. The test rejects independence
    if

    .. math::

        \frac{n \mathcal{V}_n}{T_2} \geq \mathcal{X}_{1-\alpha}^2,

    where :math:`n` is the number of samples, :math:`\mathcal{V}_n` is the
    sample distance correlation between the selected point and the target,
    :math:`\mathcal{X}_{1-\alpha}^2` is the :math:`1-\alpha` quantile of a
    chi-squared variable with 1 degree of freedom. :math:`T_2` is the product
    of the means of the distance matrices of the selected point and the
    target, a term which is involved in the standard computation of the sample
    distance covariance.

    Parameters:

        significance (float): Significance used in the independence test. By
                              default is 0.01 (1%).

    References:
        .. footbibliography::

    """

    def __init__(self, significance=0.01):
        super().__init__()
        self.significance = significance

    def chi_bound(self, x, y, significance):

        x_dist = dcor.distances.pairwise_distances(x)
        y_dist = dcor.distances.pairwise_distances(y)

        t2 = np.mean(x_dist) * np.mean(y_dist)

        chi_quant = scipy.stats.chi2.ppf(1 - significance, df=1)

        return chi_quant * t2 / x_dist.shape[0]

    def __call__(self, *, selected_variable, y, **kwargs):

        bound = self.chi_bound(selected_variable, y, self.significance)

        return dcor.u_distance_covariance_sqr(selected_variable, y) < bound


class RedundancyCondition(abc.ABC, sklearn.base.BaseEstimator):
    """
    Redundancy condition for RMH.

    This is a callable that should return ``True`` if the two points are
    redundant and false otherwise.

    """

    @abc.abstractmethod
    def __call__(self, max_point, test_point, **kwargs):
        pass


class DependenceThresholdRedundancy(RedundancyCondition):
    """
    The points are redundant if their dependency is above a given
    threshold.

    This stopping condition requires that the dependency has a known bound, for
    example that it takes values in the interval :math:`[0, 1]`.

    Parameters:

        threshold (float): Value compared with the score. If the score
            of the selected point is not higher than that,
            the point will not be selected (unless it is
            the first iteration) and RMH will end.
        dependence_measure (callable): Dependence measure to use. By default,
            it uses the bias corrected squared distance correlation.

    """

    def __init__(self, threshold=0.9, *,
                 dependence_measure=dcor.u_distance_correlation_sqr):

        super().__init__()
        self.threshold = threshold
        self.dependence_measure = dependence_measure

    def __call__(self, *, max_point, test_point, **kwargs):

        return self.dependence_measure(max_point, test_point) > self.threshold


class RMHResult(object):

    def __init__(self, index, score):
        self.index = index
        self.score = score
        self.matrix_after_correction = None
        self.original_dependence = None
        self.influence_mask = None
        self.current_mask = None

    def __repr__(self):
        return (self.__class__.__name__ +
                "(index={index}, score={score})"
                .format(index=self.index, score=self.score))


def _get_influence_mask(X, t_max_index, redundancy_condition, old_mask):
    """
    Get the mask of the points that have a large dependence with the
    selected point.

    """
    sl = slice(None)

    def get_index(index):
        return (sl,) + tuple(index) + (np.newaxis,)

    def is_redundant(index):

        max_point = np.squeeze(X[get_index(t_max_index)], axis=1)
        test_point = np.squeeze(X[get_index(index)], axis=1)

        return redundancy_condition(max_point=max_point,
                                    test_point=test_point)

    def adjacent_indexes(index):
        for i, coord in enumerate(index):
            # Out of bounds right check
            if coord < (X.shape[i + 1] - 1):
                new_index = list(index)
                new_index[i] += 1
                yield tuple(new_index)
            # Out of bounds left check
            if coord > 0:
                new_index = list(index)
                new_index[i] -= 1
                yield tuple(new_index)

    def update_mask(new_mask, index):
        indexes = [index]

        while indexes:
            index = indexes.pop()
            # Check if it wasn't masked before
            if (
                not old_mask[index] and not new_mask[index] and
                is_redundant(index)
            ):
                new_mask[index] = True
                for i in adjacent_indexes(index):
                    indexes.append(i)

    new_mask = np.zeros_like(old_mask)

    update_mask(new_mask, t_max_index)

    # The selected point is masked even if min_redundancy is high
    new_mask[t_max_index] = True

    return new_mask


def _rec_maxima_hunting_gen_no_copy(
        X: FDataGrid, y, *,
        dependence_measure=dcor.u_distance_correlation_sqr,
        correction=None,
        redundancy_condition=None,
        stopping_condition=None,
        mask=None,
        get_intermediate_results=False):
    """
    Find the most relevant features of a function using recursive maxima
    hunting. It changes the original matrix.

    Parameters:

        dependence_measure (callable): Dependence measure to use. By default,
            it uses the bias corrected squared distance correlation.
        max_features (int): Maximum number of features to select.
        correction (Correction): Correction used to subtract the information
            of each selected point in each iteration.
        redundancy_condition (callable): Condition to consider a point
            redundant with the selected maxima and discard it from future
            consideration as a maximum.
        stopping_condition (callable): Condition to stop the algorithm.
        mask (boolean array): Masked values.
        get_intermediate_results (boolean): Return additional debug info.

    """
    # X = np.asfarray(X)
    y = np.asfarray(y)

    if correction is None:
        correction = UniformCorrection()

    if mask is None:
        mask = np.zeros([len(t) for t in X.grid_points], dtype=bool)

    if redundancy_condition is None:
        redundancy_condition = DependenceThresholdRedundancy()

    if stopping_condition is None:
        stopping_condition = AsymptoticIndependenceTestStop()

    first_pass = True

    correction.begin(X, y)

    while True:
        dependences = _compute_dependence(
            X=X.data_matrix, y=y,
            dependence_measure=dependence_measure)

        t_max_index = _absolute_argmax(dependences,
                                       mask=mask)
        score = dependences[t_max_index]

        repeated_point = mask[t_max_index]

        stopping_condition_reached = stopping_condition(
            selected_index=t_max_index,
            dependences=dependences,
            selected_variable=X.data_matrix[(slice(None),) +
                                            tuple(t_max_index)],
            X=X, y=y)

        if ((repeated_point or stopping_condition_reached) and
                not first_pass):
            return

        influence_mask = _get_influence_mask(
            X=X.data_matrix, t_max_index=t_max_index,
            redundancy_condition=redundancy_condition,
            old_mask=mask)

        mask |= influence_mask

        # Correct the influence of t_max
        correction(X=X,
                   selected_index=t_max_index)
        result = RMHResult(index=t_max_index, score=score)

        # Additional info, useful for debugging
        if get_intermediate_results:
            result.matrix_after_correction = np.copy(X.data_matrix)
            result.original_dependence = dependences
            result.influence_mask = influence_mask
            result.current_mask = mask

        new_X = yield result  # Accept modifications to the matrix
        if new_X is not None:
            X.data_matrix = new_X

        correction = correction.conditioned(
            X=X.data_matrix,
            T=X.grid_points[0],
            t_0=X.grid_points[0][t_max_index])

        first_pass = False


def _rec_maxima_hunting_gen(X, *args, **kwargs):
    yield from _rec_maxima_hunting_gen_no_copy(copy.copy(X),
                                               *args, **kwargs)


class RecursiveMaximaHunting(
        sklearn.base.BaseEstimator, sklearn.base.TransformerMixin):
    r"""
    Recursive Maxima Hunting variable selection.

    This is a filter variable selection method for problems with a target
    variable. It evaluates a dependence measure between each point of the
    function and the target variable, selects the point that maximizes this
    dependence, subtracts the information of the selected point from
    the original functions and repeat the process.

    This method is inspired by :class:`MaximaHunting`, and shares
    similarities with it. However, as the information of the selected point
    is subtracted from each function in each step of the algorithm, this
    algorithm can uncover points that are not relevant by themselves but are
    relevant once other points are selected. Those points would not be
    selected by :class:`MaximaHunting` alone.

    This method was originally described in a special case in article
    :footcite:`torrecilla+suarez_2016_hunting`.
    Additional information about the usage of this method can be found in
    :doc:`/modules/preprocessing/dim_reduction/recursive_maxima_hunting`.

    Parameters:

        dependence_measure (callable): Dependence measure to use. By default,
            it uses the bias corrected squared distance correlation.
        max_features (int): Maximum number of features to select. By default
            there is no limit.
        correction (Correction): Correction used to subtract the information
            of each selected point in each iteration. By default it is
            a :class:`.UniformCorrection` object.
        redundancy_condition (callable): Condition to consider a point
            redundant with the selected maxima and discard it from future
            consideration as a maximum. By default it is a
            :class:`.DependenceThresholdRedundancy` object.
        stopping_condition (callable): Condition to stop the algorithm. By
            default it is a :class:`.AsymptoticIndependenceTestStop`
            object.

    Examples:

        >>> from skfda.preprocessing.dim_reduction import variable_selection
        >>> from skfda.datasets import make_gaussian_process
        >>> import skfda
        >>> import numpy as np

        We create trajectories from two classes, one with zero mean and the
        other with a peak-like mean. Both have Brownian covariance.

        >>> n_samples = 1000
        >>> n_features = 100
        >>>
        >>> def mean_1(t):
        ...     return (
        ...         np.abs(t - 0.25)
        ...         - 2 * np.abs(t - 0.5)
        ...         + np.abs(t - 0.75)
        ...     )
        >>>
        >>> X_0 = make_gaussian_process(
        ...     n_samples=n_samples // 2,
        ...     n_features=n_features,
        ...     random_state=0,
        ... )
        >>> X_1 = make_gaussian_process(
        ...     n_samples=n_samples // 2,
        ...     n_features=n_features,
        ...     mean=mean_1,
        ...     random_state=1,
        ... )
        >>> X = skfda.concatenate((X_0, X_1))
        >>>
        >>> y = np.zeros(n_samples)
        >>> y [n_samples // 2:] = 1

        Select the relevant points to distinguish the two classes

        >>> rmh = variable_selection.RecursiveMaximaHunting()
        >>> _ = rmh.fit(X, y)
        >>> point_mask = rmh.get_support()
        >>> points = X.grid_points[0][point_mask]
        >>> np.allclose(points, [0.25, 0.5, 0.75], rtol=1e-1)
        True

        Apply the learned dimensionality reduction

        >>> X_dimred = rmh.transform(X)
        >>> len(X.grid_points[0])
        100
        >>> X_dimred.shape
        (1000, 3)

    References:
        .. footbibliography::

    """

    def __init__(self, *,
                 dependence_measure=dcor.u_distance_correlation_sqr,
                 max_features=None,
                 correction=None,
                 redundancy_condition=None,
                 stopping_condition=None):
        self.dependence_measure = dependence_measure
        self.max_features = max_features
        self.correction = correction
        self.redundancy_condition = redundancy_condition
        self.stopping_condition = stopping_condition

    def fit(self, X, y):

        self.features_shape_ = X.data_matrix.shape[1:]

        indexes = []
        for i, result in enumerate(
            _rec_maxima_hunting_gen(
                X=X.copy(),
                y=y,
                dependence_measure=self.dependence_measure,
                correction=self.correction,
                redundancy_condition=self.redundancy_condition,
                stopping_condition=self.stopping_condition)):

            indexes.append(result.index)

            if self.max_features is not None and i >= self.max_features:
                break

        self.indexes_ = tuple(np.transpose(indexes).tolist())

        return self

    def transform(self, X):

        X_matrix = X.data_matrix

        sklearn.utils.validation.check_is_fitted(self)

        if X_matrix.shape[1:] != self.features_shape_:
            raise ValueError("The trajectories have a different number of "
                             "points than the ones fitted")

        output = X_matrix[(slice(None),) + self.indexes_]

        return output.reshape(X.n_samples, -1)

    def get_support(self, indices: bool=False):

        if indices:
            return self.indexes_
        else:
            mask = np.zeros(self.features_shape_[0], dtype=bool)
            mask[self.indexes_] = True
            return mask



"""Methods to solve least squares problems."""
from __future__ import annotations

from typing import Callable, Optional, Union

import numpy as np
from typing_extensions import Final, Literal

import scipy.linalg

LstsqMethodCallable = Callable[[np.ndarray, np.ndarray], np.ndarray]
LstsqMethodName = Literal["cholesky", "qr", "svd"]
LstsqMethod = Union[LstsqMethodCallable, LstsqMethodName]


def lstsq_cholesky(
    coefs: np.ndarray,
    result: np.ndarray,
) -> np.ndarray:
    """Solve OLS problem using a Cholesky decomposition."""
    left = coefs.T @ coefs
    right = coefs.T @ result
    return scipy.linalg.solve(left, right, assume_a="pos")


def lstsq_qr(
    coefs: np.ndarray,
    result: np.ndarray,
) -> np.ndarray:
    """Solve OLS problem using a QR decomposition."""
    return scipy.linalg.lstsq(coefs, result, lapack_driver="gelsy")[0]


def lstsq_svd(
    coefs: np.ndarray,
    result: np.ndarray,
) -> np.ndarray:
    """Solve OLS problem using a SVD decomposition."""
    return scipy.linalg.lstsq(coefs, result, lapack_driver="gelsd")[0]


method_dict: Final = {
    "cholesky": lstsq_cholesky,
    "qr": lstsq_qr,
    "svd": lstsq_svd,
}


def _get_lstsq_method(
    method: LstsqMethod,
) -> LstsqMethodCallable:
    """Convert method string to method if necessary."""
    return method if callable(method) else method_dict[method]


def solve_regularized_weighted_lstsq(
    coefs: np.ndarray,
    result: np.ndarray,
    *,
    weights: Optional[np.ndarray] = None,
    penalty_matrix: Optional[np.ndarray] = None,
    lstsq_method: LstsqMethod = lstsq_svd,
) -> np.ndarray:
    """
    Solve a regularized and weighted least squares problem.

    If the penalty matrix is not ``None`` and nonzero, there
    is a closed solution. Otherwise the problem can be reduced
    to a least squares problem.

    """
    lstsq_method = _get_lstsq_method(lstsq_method)

    if lstsq_method is not lstsq_cholesky and (
        penalty_matrix is None
    ):
        # Weighted least squares case
        if weights is not None:
            weights_chol = scipy.linalg.cholesky(weights)
            coefs = weights_chol @ coefs
            result = weights_chol @ result

        return lstsq_method(coefs, result)

    # Cholesky case (always used for the regularized case)
    if weights is None:
        left = coefs.T @ coefs
        right = coefs.T @ result
    else:
        left = coefs.T @ weights @ coefs
        right = coefs.T @ weights @ result

    if penalty_matrix is not None:
        left += penalty_matrix

    return scipy.linalg.solve(left, right, assume_a="pos")



"""Regression."""
from ._historical_linear_model import HistoricalLinearRegression
from ._linear_regression import LinearRegression
from ._neighbors_regression import (
    KNeighborsRegressor,
    RadiusNeighborsRegressor,
)



from __future__ import annotations

import enum
from typing import Callable, Generic, Optional, TypeVar, Union

import joblib
import numpy as np
import sklearn.cluster
from sklearn.base import BaseEstimator, ClusterMixin
from typing_extensions import Literal

from ...misc.metrics import PRECOMPUTED, Metric, PairwiseMetric, l2_distance
from ...misc.metrics._typing import _parse_metric, _PrecomputedTypes
from ...representation import FData

kk = ["ward", "average", "complete"]

MetricElementType = TypeVar(
    "MetricElementType",
    contravariant=True,
    bound=FData,
)

MetricOrPrecomputed = Union[Metric[MetricElementType], _PrecomputedTypes]
Connectivity = Union[
    np.ndarray,
    Callable[[MetricElementType], np.ndarray],
    None,
]


class LinkageCriterion(enum.Enum):
    """Linkage criterion to use in :class:`AgglomerativeClustering`."""

    # WARD = "ward" Not until
    # https://github.com/scikit-learn/scikit-learn/issues/15287 is solved
    COMPLETE = "complete"
    AVERAGE = "average"
    SINGLE = "single"


LinkageCriterionLike = Union[
    LinkageCriterion,
    Literal["ward", "complete", "average", "single"],
]


class AgglomerativeClustering(  # noqa: WPS230
    ClusterMixin,  # type: ignore
    BaseEstimator,  # type: ignore
    Generic[MetricElementType],
):
    r"""
    Agglomerative Clustering.

    Recursively merges the pair of clusters that minimally increases
    a given linkage distance.

    Notes:
        This class is an extension of
        :class:`sklearn.cluster.AgglomerativeClustering` that accepts
        functional data objects and metrics. Please check also the
        documentation of the original class.

    Parameters:
        n_clusters:
            The number of clusters to find. It must be ``None`` if
            ``distance_threshold`` is not ``None``.
        metric:
            Metric used to compute the linkage.
            If it is ``skfda.misc.metrics.PRECOMPUTED`` or the string
            ``"precomputed"``, a distance matrix (instead of a similarity
            matrix) is needed as input for the fit method.
        memory:
            Used to cache the output of the computation of the tree.
            By default, no caching is done. If a string is given, it is the
            path to the caching directory.
        connectivity:
            Connectivity matrix. Defines for each sample the neighboring
            samples following a given structure of the data.
            This can be a connectivity matrix itself or a callable that
            transforms the data into a connectivity matrix, such as derived
            from kneighbors_graph. Default is None, i.e, the
            hierarchical clustering algorithm is unstructured.
        compute_full_tree:
            Stop early the construction of the tree at n_clusters. This is
            useful to decrease computation time if the number of clusters
            is not small compared to the number of samples. This option is
            useful only when specifying a connectivity matrix. Note also
            that when varying the number of clusters and using caching, it
            may be advantageous to compute the full tree. It must be ``True``
            if ``distance_threshold`` is not ``None``. By default
            `compute_full_tree` is "auto", which is equivalent to `True` when
            `distance_threshold` is not `None` or that `n_clusters` is
            inferior to the maximum between 100 or `0.02 * n_samples`.
            Otherwise, "auto" is equivalent to `False`.
        linkage:
            Which linkage criterion to use. The linkage criterion determines
            which distance to use between sets of observation. The algorithm
            will merge the pairs of clusters that minimize this criterion.

            - average uses the average of the distances of each observation of
              the two sets.
            - complete or maximum linkage uses the maximum distances between
              all observations of the two sets.
            - single uses the minimum of the distances between all observations
              of the two sets.
        distance_threshold:
            The linkage distance threshold above which, clusters will not be
            merged. If not ``None``, ``n_clusters`` must be ``None`` and
            ``compute_full_tree`` must be ``True``.

    Attributes:
        n_clusters\_:
            The number of clusters found by the algorithm. If
            ``distance_threshold=None``, it will be equal to the given
            ``n_clusters``.
        labels\_:
            cluster labels for each point
        n_leaves\_:
            Number of leaves in the hierarchical tree.
        n_connected_components\_:
            The estimated number of connected components in the graph.
        children\_ :
            The children of each non-leaf node. Values less than `n_samples`
            correspond to leaves of the tree which are the original samples.
            A node `i` greater than or equal to `n_samples` is a non-leaf
            node and has children `children_[i - n_samples]`. Alternatively
            at the i-th iteration, children[i][0] and children[i][1]
            are merged to form node `n_samples + i`


    Examples:
        >>> from skfda import FDataGrid
        >>> from skfda.ml.clustering import AgglomerativeClustering
        >>> import numpy as np
        >>> data_matrix = np.array([[1, 2], [1, 4], [1, 0],
        ...                        [4, 2], [4, 4], [4, 0]])
        >>> X = FDataGrid(data_matrix)
        >>> clustering = AgglomerativeClustering(
        ...     linkage=AgglomerativeClustering.LinkageCriterion.COMPLETE,
        ... )
        >>> clustering.fit(X)
        AgglomerativeClustering(...)
        >>> clustering.labels_.astype(np.int_)
        array([0, 0, 1, 0, 0, 1])
    """

    LinkageCriterion = LinkageCriterion

    def __init__(
        self,
        n_clusters: Optional[int] = 2,
        *,
        metric: MetricOrPrecomputed[MetricElementType] = l2_distance,
        memory: Union[str, joblib.Memory, None] = None,
        connectivity: Connectivity[MetricElementType] = None,
        compute_full_tree: Union[Literal['auto'], bool] = 'auto',
        linkage: LinkageCriterionLike,
        distance_threshold: Optional[float] = None,
    ) -> None:
        self.n_clusters = n_clusters
        self.metric = metric
        self.memory = memory
        self.connectivity = connectivity
        self.compute_full_tree = compute_full_tree
        self.linkage = linkage
        self.distance_threshold = distance_threshold

    def _init_estimator(self) -> None:
        linkage = LinkageCriterion(self.linkage)

        self._estimator = sklearn.cluster.AgglomerativeClustering(
            n_clusters=self.n_clusters,
            affinity='precomputed',
            memory=self.memory,
            connectivity=self.connectivity,
            compute_full_tree=self.compute_full_tree,
            linkage=linkage.value,
            distance_threshold=self.distance_threshold,
        )

    def _copy_attrs(self) -> None:
        self.n_clusters_: int = self._estimator.n_clusters_
        self.labels_: np.ndarray = self._estimator.labels_
        self.n_leaves_: int = self._estimator.n_leaves_
        self.n_connected_components_: int = (
            self._estimator.n_connected_components_
        )
        self.children_: np.ndarray = self._estimator.children_

    def fit(  # noqa: D102
        self,
        X: MetricElementType,
        y: None = None,
    ) -> AgglomerativeClustering[MetricElementType]:

        self._init_estimator()

        metric = _parse_metric(self.metric)

        if metric is not PRECOMPUTED:
            data = PairwiseMetric(metric)(X)

        self._estimator.fit(data, y)

        self._copy_attrs()

        return self

    def fit_predict(  # noqa: D102
        self,
        X: MetricElementType,
        y: None = None,
    ) -> np.ndarray:

        self._init_estimator()

        metric = _parse_metric(self.metric)

        if metric is not PRECOMPUTED:
            data = PairwiseMetric(metric)(X)

        predicted = self._estimator.fit_predict(data, y)

        self._copy_attrs()

        return predicted



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable, Tuple, TypeVar, Union

import numpy as np
import scipy.linalg

from ..._utils import _same_domain
from ._basis import Basis

if TYPE_CHECKING:
    from .. import FDataBasis

T = TypeVar("T", bound='VectorValued')


class VectorValued(Basis):
    r"""Vector-valued basis.

    Basis for :term:`vector-valued functions <vector-valued function>`
    constructed from scalar-valued bases.

    For each dimension in the :term:`codomain`, it uses a scalar-valued basis
    multiplying each basis by the corresponding unitary vector.

    Attributes:
        domain_range (tuple): a tuple of length ``dim_domain`` containing
            the range of input values for each dimension.
        n_basis (int): number of functions in the basis.

    Examples:
        Defines a vector-valued base over the interval :math:`[0, 5]`
        consisting on the functions

        .. math::

            1 \vec{i}, t \vec{i}, t^2 \vec{i}, 1 \vec{j}, t \vec{j}

        >>> from skfda.representation.basis import VectorValued, Monomial
        >>>
        >>> basis_x = Monomial(domain_range=(0,5), n_basis=3)
        >>> basis_y = Monomial(domain_range=(0,5), n_basis=2)
        >>>
        >>> basis = VectorValued([basis_x, basis_y])


        And evaluates all the functions in the basis in a list of descrete
        values.

        >>> basis([0., 1., 2.])
        array([[[ 1.,  0.],
                [ 1.,  0.],
                [ 1.,  0.]],
               [[ 0.,  0.],
                [ 1.,  0.],
                [ 2.,  0.]],
               [[ 0.,  0.],
                [ 1.,  0.],
                [ 4.,  0.]],
               [[ 0.,  1.],
                [ 0.,  1.],
                [ 0.,  1.]],
               [[ 0.,  0.],
                [ 0.,  1.],
                [ 0.,  2.]]])

    """

    def __init__(self, basis_list: Iterable[Basis]) -> None:

        basis_list = tuple(basis_list)

        if not all(b.dim_codomain == 1 for b in basis_list):
            raise ValueError(
                "The basis functions must be scalar valued",
            )

        if any(
            b.dim_domain != basis_list[0].dim_domain
            or not _same_domain(b, basis_list[0])
            for b in basis_list
        ):
            raise ValueError(
                "The basis must all have the same domain "
                "dimension and range",
            )

        self._basis_list = basis_list

        super().__init__(
            domain_range=basis_list[0].domain_range,
            n_basis=sum(b.n_basis for b in basis_list),
        )

    @property
    def basis_list(self) -> Tuple[Basis, ...]:
        return self._basis_list

    @property
    def dim_domain(self) -> int:
        return self.basis_list[0].dim_domain

    @property
    def dim_codomain(self) -> int:
        return len(self.basis_list)

    def _evaluate(self, eval_points: np.ndarray) -> np.ndarray:
        matrix = np.zeros((self.n_basis, len(eval_points), self.dim_codomain))

        n_basis_eval = 0

        basis_evaluations = [b.evaluate(eval_points) for b in self.basis_list]

        for i, ev in enumerate(basis_evaluations):

            matrix[n_basis_eval:n_basis_eval + len(ev), :, i] = ev[..., 0]
            n_basis_eval += len(ev)

        return matrix

    def _derivative_basis_and_coefs(
        self: T,
        coefs: np.ndarray,
        order: int = 1,
    ) -> Tuple[T, np.ndarray]:

        n_basis_list = [b.n_basis for b in self.basis_list]
        indexes = np.cumsum(n_basis_list)

        coefs_per_basis = np.hsplit(coefs, indexes[:-1])

        basis_and_coefs = [
            b._derivative_basis_and_coefs(c, order=order)  # noqa: WPS437
            for b, c in zip(self.basis_list, coefs_per_basis)
        ]

        new_basis_list, new_coefs_list = zip(*basis_and_coefs)

        new_basis = type(self)(new_basis_list)
        new_coefs = np.hstack(new_coefs_list)

        return new_basis, new_coefs

    def _gram_matrix(self) -> np.ndarray:

        gram_matrices = [b.gram_matrix() for b in self.basis_list]

        return scipy.linalg.block_diag(*gram_matrices)

    def _coordinate_nonfull(
        self,
        coefs: np.ndarray,
        key: Union[int, slice],
    ) -> Tuple[Basis, np.ndarray]:

        basis_sizes = [b.n_basis for b in self.basis_list]
        basis_indexes = np.cumsum(basis_sizes)
        coef_splits = np.split(coefs, basis_indexes[:-1], axis=1)

        new_basis = self.basis_list[key]
        if not isinstance(new_basis, Basis):
            new_basis = VectorValued(new_basis)

        new_coefs = coef_splits[key]
        if not isinstance(new_coefs, np.ndarray):
            new_coefs = np.concatenate(coef_splits[key], axis=1)

        return new_basis, new_coefs

    def __repr__(self) -> str:
        """Representation of a Basis object."""
        return (
            f"{self.__class__.__name__}("
            f"basis_list={self.basis_list})"
        )

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other) and self.basis_list == other.basis_list

    def __hash__(self) -> int:
        return hash((super().__hash__(), self.basis_list))



import warnings
from typing import Any, Mapping, Optional, Tuple, Union, overload

import numpy as np
import pandas as pd
from numpy import ndarray
from pandas import DataFrame, Series
from sklearn.utils import Bunch
from typing_extensions import Literal

import rdata

from .. import FDataGrid


def _get_skdatasets_repositories() -> Any:
    import skdatasets

    repositories = getattr(skdatasets, "repositories", None)
    if repositories is None:
        repositories = skdatasets

    return repositories


def fdata_constructor(
    obj: Any,
    attrs: Mapping[Union[str, bytes], Any],
) -> FDataGrid:
    """
    Construct a :func:`FDataGrid` objet from a R `fdata` object.

    This constructor can be used in the dict passed to
    :func:`rdata.conversion.SimpleConverter` in order to
    convert `fdata` objects from the fda.usc package.

    """
    names = obj["names"]

    return FDataGrid(
        data_matrix=obj["data"],
        grid_points=obj["argvals"],
        domain_range=obj["rangeval"],
        dataset_name=names['main'][0],
        argument_names=(names['xlab'][0],),
        coordinate_names=(names['ylab'][0],),
    )


def functional_constructor(
    obj: Any,
    attrs: Mapping[Union[str, bytes], Any],
) -> FDataGrid:
    """
    Construct a :func:`FDataGrid` objet from a R `functional` object.

    This constructor can be used in the dict passed to
    :func:`rdata.conversion.SimpleConverter` in order to
    convert `functional` objects from the ddalpha package.

    """
    name = obj['name']
    args_label = obj['args']
    values_label = obj['vals']
    target = np.array(obj['labels']).ravel()
    dataf = obj['dataf']

    grid_points_set = {a for o in dataf for a in o["args"]}

    args_init = min(grid_points_set)
    args_end = max(grid_points_set)

    grid_points = np.arange(args_init, args_end + 1)

    data_matrix = np.zeros(shape=(len(dataf), len(grid_points)))

    for num_sample, o in enumerate(dataf):
        for t, x in zip(o["args"], o["vals"]):
            data_matrix[num_sample, t - args_init] = x

    return (
        FDataGrid(
            data_matrix=data_matrix,
            grid_points=grid_points,
            domain_range=(args_init, args_end),
            dataset_name=name[0],
            argument_names=(args_label[0],),
            coordinate_names=(values_label[0],),
        ),
        target,
    )


def fetch_cran(
    name: str,
    package_name: str,
    *,
    converter: Optional[rdata.conversion.Converter] = None,
    **kwargs: Any,
) -> Any:
    """
    Fetch a dataset from CRAN.

    Args:
        name: Dataset name.
        package_name: Name of the R package containing the dataset.
        converter: Object that performs the conversion of the R objects to
            Python objects.
        kwargs: Additional parameters for the function
            :func:`skdatasets.repositories.cran.fetch_dataset`.

    Returns:
        The dataset, with the R types converted to suitable Python
        types.

    """
    repositories = _get_skdatasets_repositories()

    if converter is None:
        converter = rdata.conversion.SimpleConverter({
            **rdata.conversion.DEFAULT_CLASS_MAP,
            "fdata": fdata_constructor,
            "functional": functional_constructor,
        })

    return repositories.cran.fetch_dataset(
        name,
        package_name,
        converter=converter,
        **kwargs,
    )


def _ucr_to_fdatagrid(name: str, data: np.ndarray) -> FDataGrid:
    if data.dtype == np.object_:
        data = np.array(data.tolist())

        # n_instances := data.shape[0]
        # dim_output  := data.shape[1]
        # n_points    := data.shape[2]

        data = np.transpose(data, axes=(0, 2, 1))

    grid_points = range(data.shape[1])

    return FDataGrid(data, grid_points=grid_points, dataset_name=name)


def fetch_ucr(name: str, **kwargs: Any) -> Bunch:
    """
    Fetch a dataset from the UCR.

    Args:
        name: Dataset name.
        kwargs: Additional parameters for the function
            :func:`skdatasets.repositories.ucr.fetch`.

    Returns:
        The dataset requested.

    Note:
        Functional multivariate datasets are not yet supported.

    References:
        Dau, Hoang Anh, Anthony Bagnall, Kaveh Kamgar, Chin-Chia Michael Yeh,
        Yan Zhu, Shaghayegh Gharghabi, Chotirat Ann Ratanamahatana, and
        Eamonn Keogh. «The UCR Time Series Archive».
        arXiv:1810.07758 [cs, stat], 17 de octubre de 2018.
        http://arxiv.org/abs/1810.07758.


    """
    repositories = _get_skdatasets_repositories()

    dataset = repositories.ucr.fetch(name, **kwargs)

    dataset['data'] = _ucr_to_fdatagrid(
        name=dataset['name'],
        data=dataset['data'],
    )
    dataset.pop('feature_names')

    data_test = dataset.get('data_test', None)
    if data_test is not None:
        dataset['data_test'] = _ucr_to_fdatagrid(
            name=dataset['name'],
            data=data_test,
        )

    return dataset


def _fetch_cran_no_encoding_warning(*args: Any, **kwargs: Any) -> Any:
    # Probably non thread safe
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=UserWarning,
            message="Unknown encoding. Assumed ASCII.",
        )
        return fetch_cran(*args, **kwargs)


def _fetch_elem_stat_learn(name: str) -> Any:
    return _fetch_cran_no_encoding_warning(
        name,
        "ElemStatLearn",
        version="0.1-7.1",
    )


def _fetch_ddalpha(name: str) -> Any:
    return _fetch_cran_no_encoding_warning(name, "ddalpha", version="1.3.4")


def _fetch_fda(name: str) -> Any:
    return _fetch_cran_no_encoding_warning(name, "fda", version="2.4.7")


def _fetch_fda_usc(name: str) -> Any:
    return _fetch_cran_no_encoding_warning(name, "fda.usc", version="1.3.0")


_param_descr = """
    Args:
        return_X_y: Return only the data and target as a tuple.
        as_frame: Return the data in a Pandas Dataframe or Series.
"""

_phoneme_descr = """
    These data arose from a collaboration between  Andreas Buja, Werner
    Stuetzle and Martin Maechler, and it is used as an illustration in the
    paper on Penalized Discriminant Analysis by Hastie, Buja and
    Tibshirani (1995).

    The data were extracted from the TIMIT database (TIMIT
    Acoustic-Phonetic Continuous Speech Corpus, NTIS, US Dept of Commerce)
    which is a widely used resource for research in speech recognition.  A
    dataset was formed by selecting five phonemes for
    classification based on digitized speech from this database.  The
    phonemes are transcribed as follows: "sh" as in "she", "dcl" as in
    "dark", "iy" as the vowel in "she", "aa" as the vowel in "dark", and
    "ao" as the first vowel in "water".  From continuous speech of 50 male
    speakers, 4509 speech frames of 32 msec duration were selected,
    approximately 2 examples of each phoneme from each speaker.  Each
    speech frame is represented by 512 samples at a 16kHz sampling rate,
    and each frame represents one of the above five phonemes.  The
    breakdown of the 4509 speech frames into phoneme frequencies is as
    follows:

    === ==== === ==== ===
     aa   ao dcl   iy  sh
    === ==== === ==== ===
    695 1022 757 1163 872
    === ==== === ==== ===

    From each speech frame, a log-periodogram was computed, which is one of
    several widely used methods for casting speech data in a form suitable
    for speech recognition.  Thus the data used in what follows consist of
    4509 log-periodograms of length 256, with known class (phoneme)
    memberships.

    The data contain curves sampled at 256 points, a response
    variable, and a column labelled "speaker" identifying the
    diffferent speakers.

    References:
        Hastie, Trevor; Buja, Andreas; Tibshirani, Robert. Penalized
        Discriminant Analysis. Ann. Statist. 23 (1995), no. 1, 73--102.
        doi:10.1214/aos/1176324456.
        https://projecteuclid.org/euclid.aos/1176324456
    """


@overload
def fetch_phoneme(
    *,
    return_X_y: Literal[False] = False,
    as_frame: bool = False,
) -> Bunch:
    pass


@overload
def fetch_phoneme(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[False] = False,
) -> Tuple[FDataGrid, ndarray]:
    pass


@overload
def fetch_phoneme(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[True],
) -> Tuple[DataFrame, Series]:
    pass


def fetch_phoneme(
    *,
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Union[Bunch, Tuple[FDataGrid, ndarray], Tuple[DataFrame, Series]]:
    """
    Load the phoneme dataset.

    The data is obtained from the R package 'ElemStatLearn', which takes it
    from the dataset in `https://web.stanford.edu/~hastie/ElemStatLearn/`.

    """
    descr = _phoneme_descr

    raw_dataset = _fetch_elem_stat_learn("phoneme")

    data = raw_dataset["phoneme"]

    n_points = 256

    curve_data = data.iloc[:, 0:n_points]
    sound = data["g"].values
    speaker = data["speaker"].values

    curves = FDataGrid(
        data_matrix=curve_data.values,
        grid_points=np.linspace(0, 8, n_points),
        domain_range=[0, 8],
        dataset_name="Phoneme",
        argument_names=("frequency (kHz)",),
        coordinate_names=("log-periodogram",),
    )

    curve_name = "log-periodogram"
    target_name = "phoneme"
    frame = None

    if as_frame:
        frame = pd.DataFrame({
            curve_name: curves,
            target_name: sound,
        })
        curves = frame.iloc[:, [0]]
        target = frame.iloc[:, 1]
        meta = pd.Series(speaker, name="speaker")
    else:
        target = sound.codes
        meta = np.array([speaker]).T

    if return_X_y:
        return curves, target

    return Bunch(
        data=curves,
        target=target,
        frame=frame,
        categories={target_name: sound.categories.tolist()},
        feature_names=[curve_name],
        target_names=[target_name],
        meta=meta,
        meta_names=["speaker"],
        DESCR=descr,
    )


if fetch_phoneme.__doc__ is not None:  # docstrings can be stripped off
    fetch_phoneme.__doc__ += _phoneme_descr + _param_descr

_growth_descr = """
    The Berkeley Growth Study (Tuddenham and Snyder, 1954) recorded the
    heights of 54 girls and 39 boys between the ages of 1 and 18 years.
    Heights were measured at 31 ages for each child, and the standard
    error of these measurements was about 3mm, tending to be larger in
    early childhood and lower in later years.

    References:
        Tuddenham, R. D., and Snyder, M. M. (1954) "Physical growth of
        California boys and girls from birth to age 18",
        University of California Publications in Child Development,
        1, 183-364.
"""


@overload
def fetch_growth(
    *,
    return_X_y: Literal[False] = False,
    as_frame: bool = False,
) -> Bunch:
    pass


@overload
def fetch_growth(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[False] = False,
) -> Tuple[FDataGrid, ndarray]:
    pass


@overload
def fetch_growth(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[True],
) -> Tuple[DataFrame, Series]:
    pass


def fetch_growth(
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Union[Bunch, Tuple[FDataGrid, ndarray], Tuple[DataFrame, Series]]:
    """
    Load the Berkeley Growth Study dataset.

    The data is obtained from the R package 'fda', which takes it from the
    Berkeley Growth Study.

    """
    descr = _growth_descr

    raw_dataset = _fetch_fda("growth")

    data = raw_dataset["growth"]

    ages = data["age"]
    females = data["hgtf"].T
    males = data["hgtm"].T

    sex = np.array([0] * males.shape[0] + [1] * females.shape[0])
    curves = FDataGrid(
        data_matrix=np.concatenate((males, females), axis=0),
        grid_points=ages,
        dataset_name="Berkeley Growth Study",
        argument_names=("age",),
        coordinate_names=("height",),
    )

    curve_name = "height"
    target_name = "sex"
    target_categories = ["male", "female"]
    frame = None

    if as_frame:
        sex = pd.Categorical.from_codes(sex, categories=target_categories)
        frame = pd.DataFrame({
            curve_name: curves,
            target_name: sex,
        })
        curves = frame.iloc[:, [0]]
        sex = frame.iloc[:, 1]

    if return_X_y:
        return curves, sex

    return Bunch(
        data=curves,
        target=sex,
        frame=frame,
        categories={target_name: target_categories},
        feature_names=[curve_name],
        target_names=[target_name],
        DESCR=descr,
    )


if fetch_growth.__doc__ is not None:  # docstrings can be stripped off
    fetch_growth.__doc__ += _growth_descr + _param_descr

_tecator_descr = """
    This is the Tecator data set: The task is to predict the fat content of a
    meat sample on the basis of its near infrared absorbance spectrum.

    1. Statement of permission from Tecator (the original data source)

    These data are recorded on a Tecator Infratec Food and Feed Analyzer
    working in the wavelength range 850 - 1050 nm by the Near Infrared
    Transmission (NIT) principle. Each sample contains finely chopped pure
    meat with different moisture, fat and protein contents.

    If results from these data are used in a publication we want you to
    mention the instrument and company name (Tecator) in the publication.
    In addition, please send a preprint of your article to

        Karin Thente, Tecator AB,
        Box 70, S-263 21 Hoganas, Sweden

    The data are available in the public domain with no responsability from
    the original data source. The data can be redistributed as long as this
    permission note is attached.

    For more information about the instrument - call Perstorp Analytical's
    representative in your area.


    2. Description of the data

    For each meat sample the data consists of a 100 channel spectrum of
    absorbances and the contents of moisture (water), fat and protein.
    The absorbance is -log10 of the transmittance
    measured by the spectrometer. The three contents, measured in percent,
    are determined by analytic chemistry.

    There are 215 samples.
"""


@overload
def fetch_tecator(
    *,
    return_X_y: Literal[False] = False,
    as_frame: bool = False,
) -> Bunch:
    pass


@overload
def fetch_tecator(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[False] = False,
) -> Tuple[FDataGrid, ndarray]:
    pass


@overload
def fetch_tecator(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[True],
) -> Tuple[DataFrame, DataFrame]:
    pass


def fetch_tecator(
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Union[Bunch, Tuple[FDataGrid, ndarray], Tuple[DataFrame, DataFrame]]:
    """
    Load the Tecator dataset.

    The data is obtained from the R package 'fda.usc', which takes it from
    http://lib.stat.cmu.edu/datasets/tecator.

    """
    descr = _tecator_descr

    raw_dataset = _fetch_fda_usc("tecator")

    data = raw_dataset["tecator"]

    curves = data['absorp.fdata']
    target = data['y'].rename(columns=str.lower)
    feature_name = curves.dataset_name.lower()
    target_names = target.columns.values.tolist()

    frame = None

    if as_frame:
        curves = pd.DataFrame({feature_name: curves})
        frame = pd.concat([curves, target], axis=1)
    else:
        target = target.values

    if return_X_y:
        return curves, target

    return Bunch(
        data=curves,
        target=target,
        frame=frame,
        categories={},
        feature_names=[feature_name],
        target_names=target_names,
        DESCR=descr,
    )


if fetch_tecator.__doc__ is not None:  # docstrings can be stripped off
    fetch_tecator.__doc__ += _tecator_descr + _param_descr

_medflies_descr = """
    The data set medfly1000.dat  consists of number of eggs laid daily for
    each of 1000 medflies (Mediterranean fruit flies, Ceratitis capitata)
    until time of death. Data were obtained in Dr. Carey's laboratory.
    A description of the experiment which was done by Professor Carey of
    UC Davis and collaborators in a medfly rearing facility in
    Mexico is in Carey et al.(1998) below. The main questions are to
    explore the relationship of age patterns of fecundity to mortality,
    longevity and lifetime reproduction.

    A basic finding was that individual mortality is associated with the
    time-dynamics of the egg-laying trajectory. An approximate parametric model
    of the egg laying process was developed and used in Müller et al. (2001).
    Nonparametric approaches which extend principal component analysis for
    curve data to the situation when covariates are present have been
    developed and discussed in  Chiou, Müller and Wang (2003)
    and Chiou et al. (2003).

    References:
        Carey, J.R., Liedo, P., Müller, H.G., Wang, J.L., Chiou, J.M. (1998).
        Relationship of age patterns of fecundity to mortality, longevity,
        and lifetime reproduction in a large cohort of Mediterranean fruit
        fly females. J. of Gerontology --Biological Sciences 53, 245-251.

        Chiou, J.M., Müller, H.G., Wang, J.L. (2003). Functional
        quasi-likelihood regression models with smooth random effects.
        J. Royal Statist. Soc. B65, 405-423. (PDF)

        Chiou, J.M., Müller, H.G., Wang, J.L., Carey, J.R. (2003). A functional
        multiplicative effects model for longitudinal data, with
        application to reproductive histories of female medflies.
        Statistica Sinica 13, 1119-1133. (PDF)

        Chiou, J.M., Müller, H.G., Wang, J.L. (2004).Functional response
        models. Statistica Sinica 14,675-693. (PDF)
"""


@overload
def fetch_medflies(
    *,
    return_X_y: Literal[False] = False,
    as_frame: bool = False,
) -> Bunch:
    pass


@overload
def fetch_medflies(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[False] = False,
) -> Tuple[FDataGrid, ndarray]:
    pass


@overload
def fetch_medflies(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[True],
) -> Tuple[DataFrame, Series]:
    pass


def fetch_medflies(
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Union[Bunch, Tuple[FDataGrid, ndarray], Tuple[DataFrame, Series]]:
    """
    Load the Medflies dataset.

    The data is obtained from the R package 'ddalpha', which its a modification
    of the dataset in http://www.stat.ucdavis.edu/~wang/data/medfly1000.htm.

    """
    descr = _medflies_descr

    raw_dataset = _fetch_ddalpha("medflies")

    data = raw_dataset["medflies"]

    curves = data[0]

    unique = np.unique(data[1], return_inverse=True)
    target_categories = [unique[0][1], unique[0][0]]
    target = 1 - unique[1]
    curve_name = 'eggs'
    target_name = "lifetime"

    frame = None

    if as_frame:
        target = pd.Categorical.from_codes(
            target,
            categories=target_categories,
        )
        frame = pd.DataFrame({
            curve_name: curves,
            target_name: target,
        })
        curves = frame.iloc[:, [0]]
        target = frame.iloc[:, 1]

    if return_X_y:
        return curves, target

    return Bunch(
        data=curves,
        target=target,
        frame=frame,
        categories={target_name: target_categories},
        feature_names=[curve_name],
        target_names=[target_name],
        DESCR=descr,
    )


if fetch_medflies.__doc__ is not None:  # docstrings can be stripped off
    fetch_medflies.__doc__ += _medflies_descr + _param_descr

_weather_descr = """
    Daily temperature and precipitation at 35 different locations in Canada
    averaged over 1960 to 1994.

    References:
        Ramsay, James O., and Silverman, Bernard W. (2006),
        Functional Data Analysis, 2nd ed. , Springer, New York.

        Ramsay, James O., and Silverman, Bernard W. (2002),
        Applied Functional Data Analysis, Springer, New York
"""


@overload
def fetch_weather(
    *,
    return_X_y: Literal[False] = False,
    as_frame: bool = False,
) -> Bunch:
    pass


@overload
def fetch_weather(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[False] = False,
) -> Tuple[FDataGrid, ndarray]:
    pass


@overload
def fetch_weather(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[True],
) -> Tuple[DataFrame, Series]:
    pass


def fetch_weather(
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Union[Bunch, Tuple[FDataGrid, ndarray], Tuple[DataFrame, Series]]:
    """
    Load the Canadian Weather dataset.

    The data is obtained from the R package 'fda' from CRAN.

    """
    descr = _weather_descr

    data = _fetch_fda("CanadianWeather")["CanadianWeather"]

    # Axes 0 and 1 must be transposed since in the downloaded dataset the
    # data_matrix shape is (nfeatures, n_samples, dim_codomain) while our
    # data_matrix shape is (n_samples, nfeatures, dim_codomain).
    temp_prec_daily = np.transpose(
        np.asarray(data["dailyAv"])[:, :, 0:2], axes=(1, 0, 2),
    )

    days_in_year = 365

    curves = FDataGrid(
        data_matrix=temp_prec_daily,
        grid_points=np.arange(0, days_in_year) + 0.5,
        domain_range=(0, days_in_year),
        dataset_name="Canadian Weather",
        sample_names=data["place"],
        argument_names=("day",),
        coordinate_names=(
            "temperature (ºC)",
            "precipitation (mm.)",
        ),
    )

    curve_name = "daily averages"
    target_name = "region"
    target_categories, target = np.unique(data["region"], return_inverse=True)

    frame = None

    if as_frame:
        target = pd.Categorical.from_codes(
            target,
            categories=target_categories,
        )
        frame = pd.DataFrame({
            curve_name: curves,
            "place": data["place"],
            "province": data["province"],
            "latitude": np.asarray(data["coordinates"])[:, 0],
            "longitude": np.asarray(data["coordinates"])[:, 1],
            "index": data["geogindex"],
            "monthly temperatures": np.asarray(
                data["monthlyTemp"],
            ).T.tolist(),
            "monthly precipitation": np.asarray(
                data["monthlyPrecip"],
            ).T.tolist(),
            target_name: target,
        })
        X = frame.iloc[:, :-1]
        target = frame.iloc[:, -1]
        feature_names = list(X.columns.values)

        additional_dict = {}
    else:
        feature_names = [curve_name]
        X = curves
        meta = np.concatenate(
            (
                np.array(data["place"], dtype=np.object_)[:, np.newaxis],
                np.array(data["province"], dtype=np.object_)[:, np.newaxis],
                np.asarray(data["coordinates"], dtype=np.object_),
                np.array(data["geogindex"], dtype=np.object_)[:, np.newaxis],
                np.asarray(data["monthlyTemp"]).T.tolist(),
                np.asarray(data["monthlyPrecip"]).T.tolist(),
            ),
            axis=1,
        )
        meta_names = [
            "place",
            "province",
            "latitude",
            "longitude",
            "index",
            "monthly temperatures",
            "monthly precipitation",
        ]

        additional_dict = {
            "meta": meta,
            "meta_names": meta_names,
        }

    if return_X_y:
        return X, target

    return Bunch(
        data=X,
        target=target,
        frame=frame,
        categories={target_name: target_categories},
        feature_names=feature_names,
        target_names=[target_name],
        **additional_dict,
        DESCR=descr,
    )


if fetch_weather.__doc__ is not None:  # docstrings can be stripped off
    fetch_weather.__doc__ += _weather_descr + _param_descr

_aemet_descr = """
    Series of daily summaries of 73 spanish weather stations selected for the
    period 1980-2009. The dataset contains the geographic information of each
    station and the average for the period 1980-2009 of daily temperature,
    daily precipitation and daily wind speed. Meteorological State Agency of
    Spain (AEMET), http://www.aemet.es/. Government of Spain.

    Authors:
        Manuel Febrero Bande, Manuel Oviedo de la Fuente <manuel.oviedo@usc.es>

    Source:
        The data were obtained from the FTP of AEMET in 2009.
"""


@overload
def fetch_aemet(
    *,
    return_X_y: Literal[False] = False,
    as_frame: bool = False,
) -> Bunch:
    pass


@overload
def fetch_aemet(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[False] = False,
) -> Tuple[FDataGrid, None]:
    pass


@overload
def fetch_aemet(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[True],
) -> Tuple[DataFrame, None]:
    pass


def fetch_aemet(  # noqa: WPS210
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Union[Bunch, Tuple[FDataGrid, None], Tuple[DataFrame, None]]:
    """
    Load the Spanish Weather dataset.

    The data is obtained from the R package 'fda.usc' from CRAN.

    """
    descr = _aemet_descr

    data = _fetch_fda_usc("aemet")["aemet"]

    days_in_year = 365

    data_matrix = np.empty((73, days_in_year, 3))
    data_matrix[:, :, 0] = data["temp"].data_matrix[:, :, 0]
    data_matrix[:, :, 1] = data["logprec"].data_matrix[:, :, 0]
    data_matrix[:, :, 2] = data["wind.speed"].data_matrix[:, :, 0]

    curves = data["temp"].copy(
        data_matrix=data_matrix,
        grid_points=np.arange(0, days_in_year) + 0.5,
        domain_range=(0, days_in_year),
        dataset_name="aemet",
        sample_names=data["df"].iloc[:, 1],
        argument_names=("day",),
        coordinate_names=(
            "temperature (ºC)",
            "logprecipitation",
            "wind speed (m/s)",
        ),
    )

    curve_name = "daily averages"
    df_names = [
        "index",
        "place",
        "province",
        "altitude",
        "longitude",
        "latitude",
    ]
    df_indexes = np.array([0, 1, 2, 3, 6, 7])

    frame = None

    if as_frame:
        frame = pd.DataFrame({
            curve_name: curves,
            **{
                n: data["df"].iloc[:, d]
                for (n, d) in zip(df_names, df_indexes)
            },
        })
        X = frame
        feature_names = list(X.columns.values)

        additional_dict = {}

    else:
        feature_names = [curve_name]
        X = curves
        meta = np.asarray(data["df"])[:, df_indexes]
        meta_names = df_names
        additional_dict = {
            "meta": meta,
            "meta_names": meta_names,
        }

    if return_X_y:
        return X, None

    return Bunch(
        data=X,
        target=None,
        frame=frame,
        categories={},
        feature_names=feature_names,
        **additional_dict,
        DESCR=descr,
    )


if fetch_aemet.__doc__ is not None:  # docstrings can be stripped off
    fetch_aemet.__doc__ += _aemet_descr + _param_descr


_octane_descr = """
    Near infrared (NIR) spectra of gasoline samples, with wavelengths ranging
    from 1102nm to 1552nm with measurements every two nm.
    This dataset contains six outliers to which ethanol was added, which is
    required in some states. See [RDEH2006]_ and [HuRS2015]_ for further
    details.

    The data is labeled according to this different composition.

    Source:
        Esbensen K. (2001). Multivariate data analysis in practice. 5th edn.
        Camo Software, Trondheim, Norway.

    References:
        ..  [RDEH2006] Rousseeuw, Peter & Debruyne, Michiel & Engelen, Sanne &
            Hubert, Mia. (2006). Robustness and Outlier Detection in
            Chemometrics. Critical Reviews in Analytical Chemistry. 36.
            221-242. 10.1080/10408340600969403.
        ..  [HuRS2015] Hubert, Mia & Rousseeuw, Peter & Segaert, Pieter.
            (2015). Multivariate functional outlier detection. Statistical
            Methods and Applications. 24. 177-202. 10.1007/s10260-015-0297-8.

"""


@overload
def fetch_octane(
    *,
    return_X_y: Literal[False] = False,
    as_frame: bool = False,
) -> Bunch:
    pass


@overload
def fetch_octane(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[False] = False,
) -> Tuple[FDataGrid, ndarray]:
    pass


@overload
def fetch_octane(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[True],
) -> Tuple[DataFrame, Series]:
    pass


def fetch_octane(
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Union[Bunch, Tuple[FDataGrid, ndarray], Tuple[DataFrame, Series]]:
    """Load near infrared spectra of gasoline samples.

    This function fetchs the octane dataset from the R package 'mrfDepth'
    from CRAN.

    """
    descr = _octane_descr

    # octane file from mrfDepth R package
    raw_dataset = fetch_cran("octane", "mrfDepth", version="1.0.11")
    data = raw_dataset['octane'][..., 0].T

    # The R package only stores the values of the curves, but the paper
    # describes the rest of the data. According to [RDEH2006], Section 5.4:

    # "wavelengths ranging from 1102nm to 1552nm with measurements every two
    # nm.""
    wavelength_start = 1102
    wavelength_end = 1552
    wavelength_count = 226

    grid_points = np.linspace(
        wavelength_start,
        wavelength_end,
        wavelength_count,
    )

    # "The octane data set contains six outliers (25, 26, 36–39) to which
    # alcohol was added".
    target = np.zeros(len(data), dtype=np.bool_)
    target[24:26] = 1  # noqa: WPS432
    target[35:39] = 1  # noqa: WPS432

    target_name = "is outlier"

    curve_name = "absorbances"

    curves = FDataGrid(
        data,
        grid_points=grid_points,
        dataset_name="octane",
        argument_names=("wavelength (nm)",),
        coordinate_names=("absorbances",),
    )

    frame = None

    if as_frame:
        frame = pd.DataFrame({
            curve_name: curves,
            target_name: target,
        })
        curves = frame.iloc[:, [0]]
        target = frame.iloc[:, 1]

    if return_X_y:
        return curves, target

    return Bunch(
        data=curves,
        target=target,
        frame=frame,
        categories={},
        feature_names=[curve_name],
        target_names=[target_name],
        DESCR=descr,
    )


if fetch_octane.__doc__ is not None:  # docstrings can be stripped off
    fetch_octane.__doc__ += _octane_descr + _param_descr

_gait_descr = """
    Angles formed by the hip and knee of each of 39 children over each boy
    gait cycle.

    References:
        Ramsay, James O., and Silverman, Bernard W. (2006),
        Functional Data Analysis, 2nd ed. , Springer, New York.

        Ramsay, James O., and Silverman, Bernard W. (2002),
        Applied Functional Data Analysis, Springer, New York
"""


@overload
def fetch_gait(
    *,
    return_X_y: Literal[False] = False,
    as_frame: bool = False,
) -> Bunch:
    pass


@overload
def fetch_gait(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[False] = False,
) -> Tuple[FDataGrid, None]:
    pass


@overload
def fetch_gait(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[True],
) -> Tuple[DataFrame, None]:
    pass


def fetch_gait(
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Union[Bunch, Tuple[FDataGrid, None], Tuple[DataFrame, None]]:
    """
    Load the GAIT dataset.

    The data is obtained from the R package 'fda' from CRAN.

    """
    descr = _gait_descr

    raw_data = _fetch_fda("gait")

    data = raw_data["gait"]

    data_matrix = np.asarray(data)
    data_matrix = np.transpose(data_matrix, axes=(1, 0, 2))
    grid_points = np.asarray(data.coords.get('dim_0'), np.float64)
    sample_names = np.asarray(data.coords.get('dim_1'))
    feature_name = 'gait'

    curves = FDataGrid(
        data_matrix=data_matrix,
        grid_points=grid_points,
        dataset_name=feature_name,
        sample_names=sample_names,
        argument_names=("time (proportion of gait cycle)",),
        coordinate_names=(
            "hip angle (degrees)",
            "knee angle (degrees)",
        ),
    )

    frame = None

    if as_frame:
        curves = pd.DataFrame({feature_name: curves})
        frame = curves

    if return_X_y:
        return curves, None

    return Bunch(
        data=curves,
        target=None,
        frame=frame,
        categories={},
        feature_names=[feature_name],
        target_names=[],
        DESCR=descr,
    )


if fetch_gait.__doc__ is not None:  # docstrings can be stripped off
    fetch_gait.__doc__ += _gait_descr + _param_descr

_handwriting_descr = """
    Data representing the X-Y coordinates along time obtained while
    writing the word "fda". The sample contains 20 instances measured over
    2.3 seconds that had been aligned for a better understanding. Each instance
    is formed by 1401 coordinate values.

    References:
        Ramsay, James O., and Silverman, Bernard W. (2006),
        Functional Data Analysis, 2nd ed. , Springer, New York.
"""


@overload
def fetch_handwriting(
    *,
    return_X_y: Literal[False] = False,
    as_frame: bool = False,
) -> Bunch:
    pass


@overload
def fetch_handwriting(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[False] = False,
) -> Tuple[FDataGrid, None]:
    pass


@overload
def fetch_handwriting(
    *,
    return_X_y: Literal[True],
    as_frame: Literal[True],
) -> Tuple[DataFrame, None]:
    pass


def fetch_handwriting(
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Union[Bunch, Tuple[FDataGrid, None], Tuple[DataFrame, None]]:
    """
    Load the HANDWRIT dataset.

    The data is obtained from the R package 'fda' from CRAN.

    """
    descr = _handwriting_descr

    raw_data = _fetch_fda("handwrit")

    data = raw_data["handwrit"]

    data_matrix = np.asarray(data)
    data_matrix = np.transpose(data_matrix, axes=(1, 0, 2))
    grid_points = np.asarray(data.coords.get('dim_0'), np.float64)
    sample_names = np.asarray(data.coords.get('dim_1'))
    feature_name = 'handwrit'

    curves = FDataGrid(
        data_matrix=data_matrix,
        grid_points=grid_points,
        dataset_name=feature_name,
        sample_names=sample_names,
        argument_names=("time",),
        coordinate_names=(
            "x coordinates",
            "y coordinates",
        ),
    )

    frame = None

    if as_frame:
        curves = pd.DataFrame({feature_name: curves})
        frame = curves

    if return_X_y:
        return curves, None

    return Bunch(
        data=curves,
        target=None,
        frame=frame,
        categories={},
        feature_names=[feature_name],
        target_names=[],
        DESCR=descr,
    )


if fetch_handwriting.__doc__ is not None:  # docstrings can be stripped off
    fetch_handwriting.__doc__ += _handwriting_descr + _param_descr



from typing import Tuple, TypeVar

import numpy as np
import scipy.linalg

from ._basis import Basis

T = TypeVar("T", bound='Monomial')


class Monomial(Basis):
    """Monomial basis.

    Basis formed by powers of the argument :math:`t`:

    .. math::
        1, t, t^2, t^3...

    Attributes:
        domain_range: a tuple of length 2 containing the initial and
            end values of the interval over which the basis can be evaluated.
        n_basis: number of functions in the basis.

    Examples:
        Defines a monomial base over the interval :math:`[0, 5]` consisting
        on the first 3 powers of :math:`t`: :math:`1, t, t^2`.

        >>> bs_mon = Monomial(domain_range=(0,5), n_basis=3)

        And evaluates all the functions in the basis in a list of descrete
        values.

        >>> bs_mon([0., 1., 2.])
        array([[[ 1.],
                [ 1.],
                [ 1.]],
               [[ 0.],
                [ 1.],
                [ 2.]],
               [[ 0.],
                [ 1.],
                [ 4.]]])

        And also evaluates its derivatives

        >>> deriv = bs_mon.derivative()
        >>> deriv([0, 1, 2])
        array([[[ 0.],
                [ 0.],
                [ 0.]],
               [[ 1.],
                [ 1.],
                [ 1.]],
               [[ 0.],
                [ 2.],
                [ 4.]]])
        >>> deriv2 = bs_mon.derivative(order=2)
        >>> deriv2([0, 1, 2])
        array([[[ 0.],
                [ 0.],
                [ 0.]],
               [[ 0.],
                [ 0.],
                [ 0.]],
               [[ 2.],
                [ 2.],
                [ 2.]]])
    """

    def _evaluate(self, eval_points: np.ndarray) -> np.ndarray:

        # Input is scalar
        eval_points = eval_points[..., 0]

        exps = np.arange(self.n_basis)
        raised = np.power.outer(eval_points, exps)

        return raised.T

    def _derivative_basis_and_coefs(
        self: T,
        coefs: np.ndarray,
        order: int = 1,
    ) -> Tuple[T, np.ndarray]:
        if order >= self.n_basis:
            return (
                type(self)(domain_range=self.domain_range, n_basis=1),
                np.zeros((len(coefs), 1)),
            )

        return (
            type(self)(
                domain_range=self.domain_range,
                n_basis=self.n_basis - order,
            ),
            np.array([np.polyder(x[::-1], order)[::-1] for x in coefs]),
        )

    def _gram_matrix(self) -> np.ndarray:
        integral_coefs = np.polyint(np.ones(2 * self.n_basis - 1))

        # We obtain the powers of both extremes in the domain range
        power_domain_limits = np.vander(
            self.domain_range[0], 2 * self.n_basis,
        )

        # Subtract the powers (Barrow's rule)
        power_domain_limits_diff = (
            power_domain_limits[1] - power_domain_limits[0]
        )

        # Multiply the constants that appear in the integration
        evaluated_points = integral_coefs * power_domain_limits_diff

        # Order the powers, lower to higher, discarding the constant
        # (it does not appear in the integral)
        ordered_evaluated_points = evaluated_points[-2::-1]

        # Build the matrix
        return scipy.linalg.hankel(
            ordered_evaluated_points[:self.n_basis],
            ordered_evaluated_points[self.n_basis - 1:],
        )

    def _to_R(self) -> str:  # noqa: N802
        drange = self.domain_range[0]
        rangeval = f"c({drange[0]}, {drange[1]})"
        return (
            f"create.monomial.basis("
            f"rangeval = {rangeval}, "
            f"nbasis = {self.n_basis})"
        )



"""fda package.

It includes the modules:
    - basis: For functional data manipulation in a function basis system.
    - grid: For functional data manipulation as a discrete set of measures.
    - math: Mean, variance, covariance, logarithms, square roots, distances...
    - kernels: kernels for kernel smoothing.
    - kernel_smoothers: kernel smoothers for smoothing FDataGrid objects.
    - validation: cross validation methods for finding the parameter that
    best smooths a FDataGrid object.
    - depth_measures: depth methods to order he samples of FDataGrid objects.
    - magnitude-shape plot: visualization tool.
    - fdata_boxplot: informative exploratory tool for visualizing functional
    data.
    - clustering: K-Means and Fuzzy-KMeans algorithms implemented to cluster
    data in the FDataGrid, along with plotting methods.
and the following classes:
    - FDataGrid: Discrete representation of functional data.
    - FDataBasis: Basis representation for functional data.
    - Boxplot: Implements the functional boxplot for FDataGrid with domain
    dimension 1.
    - SurfaceBoxplot: Implements the functional boxplot for FDataGrid with
    domain dimension 2.
    - MagnitudeShapePlot: Implements the magnitude shape plot for FDataGrid
    with domain dimension 1.
    -Kmeans: Implements the K-Means clustering algorithm.
    -FuzzyKmeans: Implements the Fuzzy K-Means clustering algorithm.

"""
import errno as _errno

from .representation import FData
from .representation import FDataBasis
from .representation import FDataGrid
from .representation._functional_data import concatenate

from . import representation, datasets, preprocessing, exploratory, misc, ml, \
    inference

import os as _os

try:
    with open(_os.path.join(_os.path.dirname(__file__),
                            '..', 'VERSION'), 'r') as version_file:
        __version__ = version_file.read().strip()
except IOError as e:
    if e.errno != _errno.ENOENT:
        raise

    __version__ = "0.0"



"""Classification."""
from ._centroid_classifiers import DTMClassifier, NearestCentroid
from ._depth_classifiers import (
    DDClassifier,
    DDGClassifier,
    MaximumDepthClassifier,
)
from ._logistic_regression import LogisticRegression
from ._neighbors_classifiers import (
    KNeighborsClassifier,
    RadiusNeighborsClassifier,
)



"""Magnitude-Shape Plot Module.

This module contains the necessary functions to construct the Magnitude-Shape
Plot. First the directional outlingness is calculated and then, an outliers
detection method is implemented.

"""
from __future__ import annotations

from typing import Any, Optional, Sequence, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.colors import Colormap
from matplotlib.figure import Figure
from matplotlib.patches import Ellipse

from ... import FDataGrid
from ...representation._typing import NDArrayFloat, NDArrayInt
from ..depth import Depth
from ..outliers import MSPlotOutlierDetector
from ._baseplot import BasePlot


class MagnitudeShapePlot(BasePlot):
    r"""
    Implementation of the magnitude-shape plot.

    This plot, which is based on the calculation of the :func:`directional
    outlyingness <fda.magnitude_shape_plot.directional_outlyingness>`
    of each of the samples, serves as a visualization tool for the centrality
    of curves. Furthermore, an outlier detection procedure is included.

    The norm of the mean of the directional outlyingness (:math:`\lVert
    \mathbf{MO}\rVert`) is plotted in the x-axis, and the variation of the
    directional outlyingness (:math:`VO`) in the y-axis.

    The outliers are detected using an instance of
    :class:`MSPlotOutlierDetector`.

    For more information see :footcite:ts:`dai+genton_2018_visualization`.

    Args:
        fdata (FDataGrid): Object containing the data.
        multivariate_depth (:ref:`depth measure <depth-measures>`, optional):
            Method used to order the data. Defaults to :class:`projection
            depth <fda.depth_measures.multivariate.ProjectionDepth>`.
        pointwise_weights (array_like, optional): an array containing the
            weights of each points of discretisation where values have
            been recorded.
        alpha (float, optional): Denotes the quantile to choose the cutoff
            value for detecting outliers Defaults to 0.993, which is used
            in the classical boxplot.
        assume_centered (boolean, optional): If True, the support of the
            robust location and the covariance estimates is computed, and a
            covariance estimate is recomputed from it, without centering
            the data. Useful to work with data whose mean is significantly
            equal to zero but is not exactly zero. If False, default value,
            the robust location and covariance are directly computed with
            the FastMCD algorithm without additional treatment.
        support_fraction (float, 0 < support_fraction < 1, optional): The
            proportion of points to be included in the support of the
            raw MCD estimate.
            Default is None, which implies that the minimum value of
            support_fraction will be used within the algorithm:
            [n_sample + n_features + 1] / 2
        random_state (int, RandomState instance or None, optional): If int,
            random_state is the seed used by the random number generator;
            If RandomState instance, random_state is the random number
            generator; If None, the random number generator is the
            RandomState instance used by np.random. By default, it is 0.
        ellipsoid: Whether to draw the non outlying ellipsoid.

    Attributes:
        points(numpy.ndarray): 2-dimensional matrix where each row
            contains the points plotted in the graph.
        outliers (1-D array, (fdata.n_samples,)): Contains 1 or 0 to denote
            if a sample is an outlier or not, respecively.
        colormap(matplotlib.pyplot.LinearSegmentedColormap, optional): Colormap
            from which the colors of the plot are extracted. Defaults to
            'seismic'.
        color (float, optional): Tone of the colormap in which the nonoutlier
            points are  plotted. Defaults to 0.2.
        outliercol (float, optional): Tone of the colormap in which the
            outliers are plotted. Defaults to 0.8.
        xlabel (string, optional): Label of the x-axis. Defaults to 'MO',
            mean of the  directional outlyingness.
        ylabel (string, optional): Label of the y-axis. Defaults to 'VO',
            variation of the  directional outlyingness.
        title (string, optional): Title of the plot. defaults to 'MS-Plot'.

    Representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.datasets import make_gaussian_process
        from skfda.misc.covariances import Exponential
        from skfda.exploratory.visualization import MagnitudeShapePlot

        fd = make_gaussian_process(
                n_samples=20, cov=Exponential(), random_state=1)

        MagnitudeShapePlot(fd)

    Example:
        >>> import skfda
        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> grid_points = [ 0., 2., 4., 6., 8., 10.]
        >>> fd = skfda.FDataGrid(data_matrix, grid_points)
        >>> MagnitudeShapePlot(fd)
        MagnitudeShapePlot(
            fdata=FDataGrid(
                array([[[ 1. ],
                        [ 1. ],
                        [ 2. ],
                        [ 3. ],
                        [ 2.5],
                        [ 2. ]],
                       [[ 0.5],
                        [ 0.5],
                        [ 1. ],
                        [ 2. ],
                        [ 1.5],
                        [ 1. ]],
                       [[-1. ],
                        [-1. ],
                        [-0.5],
                        [ 1. ],
                        [ 1. ],
                        [ 0.5]],
                       [[-0.5],
                        [-0.5],
                        [-0.5],
                        [-1. ],
                        [-1. ],
                        [-1. ]]]),
                grid_points=(array([  0.,   2.,   4.,   6.,   8.,  10.]),),
                domain_range=((0.0, 10.0),),
                ...),
            multivariate_depth=None,
            pointwise_weights=None,
            alpha=0.993,
            points=array([[ 1.66666667,  0.12777778],
                          [ 0.        ,  0.        ],
                          [-0.8       ,  0.17666667],
                          [-1.74444444,  0.94395062]]),
            outliers=array([False, False, False, False]),
            colormap=seismic,
            color=0.2,
            outliercol=0.8,
            xlabel='MO',
            ylabel='VO',
            title='MS-Plot')

    References:
        .. footbibliography::

    """

    def __init__(
        self,
        fdata: FDataGrid,
        chart: Union[Figure, Axes, None] = None,
        *,
        fig: Optional[Figure] = None,
        axes: Optional[Sequence[Axes]] = None,
        ellipsoid: bool = True,
        **kwargs: Any,
    ) -> None:

        BasePlot.__init__(
            self,
            chart,
            fig=fig,
            axes=axes,
        )
        if fdata.dim_codomain > 1:
            raise NotImplementedError(
                "Only support 1 dimension on the codomain.")

        self.outlier_detector = MSPlotOutlierDetector(**kwargs)

        y = self.outlier_detector.fit_predict(fdata)

        outliers = (y == -1)

        self.ellipsoid = ellipsoid

        self._fdata = fdata
        self._outliers = outliers
        self._colormap = plt.cm.get_cmap('seismic')
        self._color = 0.2
        self._outliercol = 0.8
        self.xlabel = 'MO'
        self.ylabel = 'VO'
        self.title = 'MS-Plot'

    @property
    def fdata(self) -> FDataGrid:
        return self._fdata

    @property
    def multivariate_depth(self) -> Optional[Depth[NDArrayFloat]]:
        return self.outlier_detector.multivariate_depth

    @property
    def pointwise_weights(self) -> Optional[NDArrayFloat]:
        return self.outlier_detector.pointwise_weights

    @property
    def alpha(self) -> float:
        return self.outlier_detector.alpha

    @property
    def points(self) -> NDArrayFloat:
        return self.outlier_detector.points_

    @property
    def outliers(self) -> NDArrayInt:
        return self._outliers

    @property
    def colormap(self) -> Colormap:
        return self._colormap

    @colormap.setter
    def colormap(self, value: Colormap) -> None:
        if not isinstance(value, matplotlib.colors.Colormap):
            raise ValueError(
                "colormap must be of type "
                "matplotlib.colors.Colormap",
            )
        self._colormap = value

    @property
    def color(self) -> float:
        return self._color

    @color.setter
    def color(self, value: float) -> None:
        if value < 0 or value > 1:
            raise ValueError(
                "color must be a number between 0 and 1.")

        self._color = value

    @property
    def outliercol(self) -> float:
        return self._outliercol

    @outliercol.setter
    def outliercol(self, value: float) -> None:
        if value < 0 or value > 1:
            raise ValueError(
                "outcol must be a number between 0 and 1.")
        self._outliercol = value

    @property
    def n_samples(self) -> int:
        return self.fdata.n_samples

    def _plot(
        self,
        fig: Figure,
        axes: Axes,
    ) -> None:

        self.artists = np.zeros(
            (self.n_samples, 1),
            dtype=Artist,
        )
        colors = np.zeros((self.fdata.n_samples, 4))
        colors[np.where(self.outliers == 1)] = self.colormap(self.outliercol)
        colors[np.where(self.outliers == 0)] = self.colormap(self.color)

        colors_rgba = [tuple(i) for i in colors]

        if self.ellipsoid:
            center = self.outlier_detector.cov_.location_
            prec = self.outlier_detector.cov_.get_precision()

            K = (
                self.outlier_detector.cutoff_value_
                / self.outlier_detector.scaling_
            )

            eigvals, eigvecs = np.linalg.eigh(prec)

            a, b = np.sqrt(K / eigvals)
            if eigvecs[0, 1] * eigvecs[1, 0] > 0:
                eigvecs[:, 0] *= -1

            angle = np.rad2deg(np.arctan2(eigvecs[1, 0], eigvecs[0, 0]))

            ellipse = Ellipse(
                xy=center,
                width=2 * a,
                height=2 * b,
                angle=angle,
                facecolor='C0',
                alpha=0.1,
            )

            axes[0].add_patch(ellipse)

        for i, _ in enumerate(self.points[:, 0].ravel()):
            self.artists[i, 0] = axes[0].scatter(
                self.points[:, 0].ravel()[i],
                self.points[:, 1].ravel()[i],
                color=colors_rgba[i],
                picker=True,
                pickradius=2,
            )

        axes[0].set_xlabel(self.xlabel)
        axes[0].set_ylabel(self.ylabel)
        axes[0].set_title(self.title)

    def __repr__(self) -> str:
        """Return repr(self)."""
        return (
            f"MagnitudeShapePlot("
            f"\nfdata={repr(self.fdata)},"
            f"\nmultivariate_depth={self.multivariate_depth},"
            f"\npointwise_weights={repr(self.pointwise_weights)},"
            f"\nalpha={repr(self.alpha)},"
            f"\npoints={repr(self.points)},"
            f"\noutliers={repr(self.outliers)},"
            f"\ncolormap={self.colormap.name},"
            f"\ncolor={repr(self.color)},"
            f"\noutliercol={repr(self.outliercol)},"
            f"\nxlabel={repr(self.xlabel)},"
            f"\nylabel={repr(self.ylabel)},"
            f"\ntitle={repr(self.title)})"
        ).replace('\n', '\n    ')



"""Feature extraction union for dimensionality reduction."""
from __future__ import annotations

from typing import Union

from numpy import ndarray
from pandas import DataFrame
from sklearn.pipeline import FeatureUnion

from ....representation import FData


class FDAFeatureUnion(FeatureUnion):  # type: ignore
    """Concatenates results of multiple functional transformer objects.

    This estimator applies a list of transformer objects in parallel to the
    input data, then concatenates the results (They can be either FDataGrid
    and FDataBasis objects or multivariate data itself).This is useful to
    combine several feature extraction mechanisms into a single transformer.
    Parameters of the transformers may be set using its name and the parameter
    name separated by a '__'. A transformer may be replaced entirely by
    setting the parameter with its name to another transformer,
    or removed by setting to 'drop'.

    Parameters:
        transformer_list: list of tuple
            List of tuple containing `(str, transformer)`. The first element
            of the tuple is name affected to the transformer while the
            second element is a scikit-learn transformer instance.
            The transformer instance can also be `"drop"` for it to be
            ignored.
        n_jobs: int
            Number of jobs to run in parallel.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
            context.
            ``-1`` means using all processors.
            The default value is None
        transformer_weights: dict
            Multiplicative weights for features per transformer.
            Keys are transformer names, values the weights.
            Raises ValueError if key not present in ``transformer_list``.
        verbose: bool
            If True, the time elapsed while fitting each transformer will be
            printed as it is completed. By default the value is False
        array_output: bool
            indicates if the transformed data is requested to be a NumPy array
            output. By default the value is False.

    Examples:
    Firstly we will import the Berkeley Growth Study data set
    >>> from skfda.datasets import fetch_growth
    >>> X,y = fetch_growth(return_X_y=True)

    Then we need to import the transformers we want to use. In our case we
    will use Generalized depth-versus-depth transformer.
    Evaluation Transformer returns the original curve, and as it is helpful,
    we will concatenate it to the already metioned transformer.
    >>> from skfda.preprocessing.dim_reduction.feature_extraction import (
    ...     FDAFeatureUnion,
    ... )
    >>> from skfda.preprocessing.dim_reduction.feature_extraction import (
    ...     DDGTransformer,
    ... )
    >>> from skfda.exploratory.depth import ModifiedBandDepth
    >>> from skfda.representation import EvaluationTransformer
    >>> import numpy as np

    Finally we apply fit and transform.
    >>> union = FDAFeatureUnion(
    ...     [
    ...        (
    ...             'ddgtransformer',
    ...             DDGTransformer(depth_method=[ModifiedBandDepth()]),
    ...        ),
    ...        ("eval", EvaluationTransformer()),
    ...     ],
    ...     array_output=True,
    ... )
    >>> np.around(union.fit_transform(X,y), decimals=2)
      array([[ 2.100e-01,  9.000e-02,  8.130e+01, ...,  1.938e+02,  1.943e+02,
               1.951e+02],
             [ 4.600e-01,  3.800e-01,  7.620e+01, ...,  1.761e+02,  1.774e+02,
               1.787e+02],
             [ 2.000e-01,  3.300e-01,  7.680e+01, ...,  1.709e+02,  1.712e+02,
               1.715e+02],
             ...,
             [ 3.900e-01,  5.100e-01,  6.860e+01, ...,  1.660e+02,  1.663e+02,
               1.668e+02],
             [ 2.600e-01,  2.700e-01,  7.990e+01, ...,  1.683e+02,  1.684e+02,
               1.686e+02],
             [ 3.300e-01,  3.200e-01,  7.610e+01, ...,  1.686e+02,  1.689e+02,
               1.692e+02]])
    """

    def __init__(
        self,
        transformer_list: list,  # type: ignore
        *,
        n_jobs: int = 1,
        transformer_weights: dict = None,  # type: ignore
        verbose: bool = False,
        array_output: bool = False,
    ) -> None:
        self.array_output = array_output
        super().__init__(
            transformer_list,
            n_jobs=n_jobs,
            transformer_weights=transformer_weights,
            verbose=verbose,
        )

    def _hstack(self, Xs: ndarray) -> Union[DataFrame, ndarray]:

        if self.array_output:
            for i in Xs:
                if isinstance(i, FData):
                    raise TypeError(
                        "There are transformed instances of FDataGrid or "
                        "FDataBasis that can't be concatenated on a NumPy "
                        "array.",
                    )
            return super()._hstack(Xs)

        return DataFrame({'Transformed data': Xs})
